package sieve.phys.circuit;

import java.awt.*;
import java.awt.event.*;
import sieve.workbench.*;
import java.io.Serializable;

public class CircuitComponentLink extends Component implements Link, 
    MouseListener {
    private Workbench wb = null;
    private Terminal sourceTerminal = null;
    private Terminal destTerminal = null;
    private UniqueID id = null;

    private CircuitComponent sourceComponent = null;
    private CircuitComponent destComponent = null;

    private Point lastPoint = null;
    private Point start = null;
    private Point end = null;

    public CircuitComponentLink(Workbench wb, CircuitComponent sourceComponent,
	Terminal sourceTerminal) {
  this.sourceComponent = sourceComponent;
  this.sourceTerminal = sourceTerminal;
	init(wb);
    }

    private void init(Workbench wb) {
	this.wb = wb;
	this.id = new UniqueID(this);
	addMouseListener(this);
	setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
    }

    public CircuitComponentLink(){}

    public void init(Workbench wb, WorkbenchComponent source,
           WorkbenchComponent dest, Serializable state) throws LinkException {

	if (! (source instanceof CircuitComponent))
	    throw new LinkException("Incompatible source component");
	if (! (dest instanceof CircuitComponent))
	    throw new LinkException("Incompatible destination component");
	if (! (state instanceof CircuitComponentLinkInfo))
	    throw new LinkException("Bad state info");

//	CircuitComponent sourceComponent = (CircuitComponent)source;
//	CircuitComponent destComponent = (CircuitComponent)dest;

	this.sourceTerminal = sourceComponent.getTerminalWithID(
	    ((CircuitComponentLinkInfo)state).getSourceTerminalID());
	this.destTerminal = destComponent.getTerminalWithID(
	    ((CircuitComponentLinkInfo)state).getDestTerminalID());
	bothTerminalsFound();

	init(wb);
    }

	    private void bothTerminalsFound() {
		sourceTerminal.addLink(this);
        	destTerminal.addLink(this);
	    }

	    public Serializable getState() {
		return new CircuitComponentLinkInfo(sourceTerminal.getID(),
		    destTerminal.getID());
	    }
	 
	    public Component getView() {
		return this;
	    }

	    /**
	     * Returns the unique ID associated with this link.
	     */
	    public UniqueID getUniqueID() {
		return id;
	    }
	 
	    /**
	     * Sets the unique ID associated with this link.
	     */
	    public void setUniqueID(UniqueID id) {
		this.id = id;
	    }

	    /**
	     * Return the component from which the link was started.
	     */
	    public WorkbenchComponent getSourceComponent() {
		if (sourceTerminal != null)
		   return sourceTerminal.getCircuitComponent();
		else
		   return null;
	    }
	 
	    /**
	     * Return the component to which the link was drawn.
	     */
	    public WorkbenchComponent getDestComponent() {
		if (destTerminal != null)
		   return destTerminal.getCircuitComponent();
		else
		   return null;
	    }

	    /**
	     * Return the terminal from which the link was started.
	     */
	    public Terminal getSourceTerminal() {
		return sourceTerminal;
	    }

	    /**
	     * Return the terminal to which the link was drawn.
	     */
	    public Terminal getDestTerminal() {
		return destTerminal;
	    }
	 
	    /**
	     * Called by the workbench when a component requests selection
	     * when the link is being created. 
	     */
	    public void setDestComponent (WorkbenchComponent c) throws LinkException {
		if (! (c instanceof CircuitComponent)) {
		    remove();
		    throw new LinkException("Destination incompatible");
		}
		if (c == null) {
		    remove();
		    throw new LinkException("Null destination component");
		}

		CircuitComponent cc = (CircuitComponent) c;

		Terminal dt = cc.getSelectedTerminal();
		if (dt == null) {
		    remove();
		    throw new LinkException("No destination terminal");
		}
		if (dt == sourceTerminal) {
		    remove();
		    throw new 
			LinkException("Attempting to connect component to itself");
		}

		destTerminal = dt;
		bothTerminalsFound();
	repaint();
    }

    public void mouseMoved(MouseEvent e) {
	lastPoint = e.getPoint();
	repaint();
    }

    public void mouseDragged(MouseEvent e) {
    }

    public void mouseClicked(MouseEvent e) {
	remove();
    }
    public  void mousePressed(MouseEvent e) { }
    public  void mouseReleased(MouseEvent e) { }
    public  void mouseEntered(MouseEvent e) { }
    public  void mouseExited(MouseEvent e) { }

    /**
     * Destroys the link. 
     */
    public void removelinks(Terminal term) {
	if ((getSourceComponent() != null) && (getDestComponent() != null)) {
	  if (sourceTerminal!=term)
              sourceTerminal.removeLink(this);
	  else
             destTerminal.removeLink(this);
	}
    }
    public void remove() {
	wb.remove(this);
	if ((getSourceComponent() != null) && (getDestComponent() != null)) {
	    wb.linkRemoved(this);
	    sourceTerminal.removeLink(this);
	    destTerminal.removeLink(this);
	}
	wb.repaint();
    }

    public void paint(Graphics g) {
	CircuitComponent sourceComponent = 
	    (CircuitComponent) getSourceComponent();
	CircuitComponent destComponent = 
	    (CircuitComponent) getDestComponent();
	if (sourceComponent != null) {
	    int x1 = sourceComponent.getLocation().x + 
	  	sourceTerminal.getLocation().x + 
		sourceTerminal.getBounds().width/2;
	    int y1 = sourceComponent.getLocation().y + 
	  	sourceTerminal.getLocation().y + 
		sourceTerminal.getBounds().height/2;
	    start = new Point(x1, y1);
	    g.setColor(Color.black);
            if (destComponent != null) {
	        int x2 = destComponent.getLocation().x + 
	  	    destTerminal.getLocation().x + 
		    destTerminal.getBounds().width/2;
	    	int y2 = destComponent.getLocation().y + 
	  	    destTerminal.getLocation().y + 
		    destTerminal.getBounds().height/2;
	        end = new Point(x2, y2);
	        g.drawLine(x1, y1, x2, y2);
	    }
	    else if (lastPoint != null) 
	        g.drawLine(x1, y1, lastPoint.x, lastPoint.y);
	}
    }

    public boolean contains(int x, int y) {
	if ((start == null) || (end == null)) 
	    return false;
	else {
	    double length = Math.sqrt((start.x - end.x) * (start.x - end.x) +
                (start.y - end.y) * (start.y - end.y));
	    double length1 = Math.sqrt((start.x - x) * (start.x - x) +
                (start.y - y) * (start.y - y));
            double length2 = Math.sqrt((end.x - x) * (end.x - x) +
                (end.y - y) * (end.y - y));
            return (Math.abs(length - length1 - length2) < 0.25);
	}
    }

}

class CircuitComponentLinkInfo implements Serializable {
    private Integer sourceTerminalID = null;
    private Integer destTerminalID = null;

    public CircuitComponentLinkInfo(Integer sourceTerminalID,
	Integer destTerminalID) {
	this.sourceTerminalID = sourceTerminalID;
	this.destTerminalID = destTerminalID;
    }
    public Integer getSourceTerminalID() {
	return sourceTerminalID;
    }
    public Integer getDestTerminalID() {
	return destTerminalID;
    }
}

