package sieve.sluice;

import java.util.*;
import java.awt.*;
import java.io.Serializable;

/**
 * Provides information about a single field in a table.
 * This should probably implement clonable.
 *
 * Add hierarchical data hint? (e.g., paths, domain names)
 *
 * Implement FieldDescriptorSupport class for filters to use.
 */
public class FieldDescriptor implements Cloneable, Serializable {
  /**
   * Field values represent labels.
   */
  public static final int LEVEL_LABEL = 0;

  /**
   * Field values describe attributes that do not imply magnitude
   * or order (e.g., race, religion, marital status)
   */
  public static final int LEVEL_NOMINAL = LEVEL_LABEL + 1;

  /**
   * Field values are ranked from highest to lowest but do not
   * imply magnitude (e.g., a ranking from shortest to tallest, but
   * without actual height data)
   */
  public static final int LEVEL_ORDINAL = LEVEL_LABEL + 2;

  /**
   * Field values are ordered and have measurable differences
   * (e.g., physically measurable attributes like length, velocity, and
   * temporature). Magnitude is implied. Values of this kind should
   * be numeric, and this should be the default kind for numeric
   * values.
   */
  public static final int LEVEL_INTERVAL = LEVEL_LABEL + 3;

  /**
   * Field values can be ordered, have measurable differences, and
   * have a zero starting point, which implies an absence of variable.
   * (e.g., age, where zero years reflects the absence of age)
   * Values of this kind should be numeric.
   */
  public static final int LEVEL_RATIO = LEVEL_LABEL + 4;

  /**
   * Hint specifying units in which this data is represented.
   */
  public static final String HINT_UNITS = "HINT_UNITS";

  protected String 	name 	= null;
  protected String 	type 	= "java.lang.Object";
  protected int 	level	= LEVEL_LABEL;
  protected String	description = null;
  protected Color	color	= Color.black;

  protected Hashtable	  labels	= null;

  // Probably don't need both of these...
  protected Hashtable	  hints	        = null;
  protected Hashtable     attributes    = null;

  public FieldDescriptor(String name) {
    this.name = name;
  }

  /**
   * Return the name of the field.
   */
  public String getName() {
    return name;
  }
  public void setName(String name) {
    this.name = name;
  }

  /**
   * Return the type (full class name) of the field.
   */
  public String getType() {
    return type;
  }
  /**
   * Set the type (full class name) of the field.
   */
  public void setType(String type) {
    this.type = type;
  }

  /**
   * Return the level of the field.
   */
  public int getLevel() {
    return level;
  }
  /**
   * Set the level of the field.
   */
  public void setLevel(int level) {
    this.level = level;
  }

  /**
   * Return the table of value labels for the field.
   */
  public Hashtable getLabels() {
    return labels;
  }
  /**
   * Set the table of value labels for the field.
   */
  public void setLabels(Hashtable labels) {
    this.labels = labels;
  }

  /**
   * Return the table of value labels for the field.
   */
  public Hashtable getAttributes() {
    return attributes;
  }
  /**
   * Set the table of value labels for the field.
   */
  public void setAttributes(Hashtable attrs) {
      this.attributes = attrs;
  }

  /**
   * Returns a (longer) description of a field. While the field name
   * need not be human-readable, the string returned by this method
   * should be. If the values in the field have a unit, the description
   * should include that as well.
   */
  public String getDescription() {
    if (description == null)
      return getName();
    else
      return description;
  }

  /**
   * Set the description of this field.
   */
  public void setDescription(String description) {
    this.description = description;
  }

  /**
   * Return the preferred display color for this field.
   */
  public Color getColor() {
    return color;
  }
  /**
   * Set the preferred display color for this field.
   */
  public void setColor(Color color) {
    this.color = color;
  }

  public FieldDescriptor copy() {
    FieldDescriptor fd = new FieldDescriptor(name);
    fd.type = type;
    fd.description = description;
    fd.level = level;
    fd.labels = labels; // Clone this?
    fd.color = color;
    fd.hints = hints; // Clone this?
    return fd;
  }

  /**
   * Copies the argument's values into this field descriptor. 
   * (Like cloning, only from within an instance.)
   */
  public void copy(FieldDescriptor fd) {
    type = fd.type;
    description = fd.description;
    level = fd.level;
    labels = fd.labels; // Clone this?
    color = fd.color;
    hints = fd.hints; // Clone this?
  }

}



