package sieve.sluice;
import java.util.*;

/**
 * Any object which implements the <CODE>TableView</CODE> interface
 * represents a "view" of a table. 
 */
public interface TableView extends TableViewListener {
    /**
     * Returns a reference to the original table from which this view was
     * constructed. This method returns a table view rather than a table
     * for several reasons:
     * <OL>
     * <LI> Table objects should probably only manipulated by 
     * thier creators. If the caller does, in fact, have a good reason
     * for manipulating an actual table, they can see if the thing
     * returned by this method is an instance of Table and cast it
     * appropriately.
     * <LI> Since the envisioned use of this method is to allow 
     * the caller to get an entire unfiltered record, there may be
     * cases where the thing returned cannot actually be a full table.
     * For example, consider a "stack" of TableView objects which
     * includes a SampleView. Since the SampleView reduces the number
     * of records in the table, the object returned by this method
     * must preserve this size reduction so that record numbers will
     * still match.
     * <LI> It may be useful to allow non-Table objects to appear
     * at the bottom of a "stack" of TableViews, particularly since Table will
     * need to inherit from our container class.
     * </OL>
     * Do we also want to be able to get the view from which we 
     * were constructed? 
     */

    public TableView table() throws NoSourceException;

    /**
     * Returns the TableDescriptor for this view's representation of
     * the table.
     */
    public TableDescriptor getTableDescriptor() throws NoSourceException;

    /**
     * Returns the names of fields in this view. These names should be
     * used to reference fields in the <CODE>get</CODE> methods.
     */
    public String[] getFieldNames() throws NoSourceException;

    /**
     * Returns the FieldDescriptor for the specified field.
     */
    public FieldDescriptor getFieldDescriptor(String field) 
        throws InterchangeException;

    /**
     * Returns number of records in the view.
     */
    public int getRecordCount() throws NoSourceException;

    /**
     * Returns no more than <CODE>numberOfRecords</CODE> records starting
     * at record number <CODE>startingAt</CODE> from 
     * the fields specified in <CODE>fields</CODE>. The resulting array
     * is ordered by <CODE>[field][record]</CODE>.
     */
    public Object[][] get(String[] fields, int startingAt, 
	int numberOfRecords) throws InterchangeException;

    /**
     * Returns all records starting from record number <CODE>startingAt</CODE>
     * from the fields specified in <CODE>fields</CODE>. (I still don't
     * like this because it doesn't add anything. _Every_ implementation
     * will almost certainly just have a call to the three-parm get...)
     */
    public Object[][] get(String[] fields, int startingAt) 
	throws InterchangeException;

    /* -------------  Event-related methods ------------- */
    public void addTableViewListener(TableViewListener listener);
    public void removeTableViewListener(TableViewListener listener);

    /* -------------  Methods for link support ------------- */
    public void addSourceView(TableView source) throws TooManySourcesException;
    public void removeSourceView(TableView source);
}

