package sieve.sluice;

import java.net.*;
import java.util.*;
import java.io.*;
import java.applet.*;

/**
 * URLFinder is a singleton object which finds a URL based on a location
 * string. It is intended to resolve conflicts over <I>where</I> tables 
 * may be loaded from. It attempts to account for numerous problems with 
 * loading data files in a variety of environments:
 * <UL>
 * <LI> Loading data files as resources from a jar file works under
 *      appletviewer (and when running in an application) but is
 *      broken on both Netscape and IE as of October 1997.
 *
 * <LI> Absolute URLs are problematic because we want to be able to use
 *      our software off-line. (Jar files would be the best solution to
 *      this problem if they were's broken.)
 *
 * <LI> Relative URLs are problematic if they reference directories other
 *      than the "current directory" (i.e., the location of the 
 *      page containing the applet) when accessing the files locally.
 *      Netscape complains that the codebase does not include any directories
 *      other than the "current directory" and throws a security exception.
 *      Note that this is only a problem when accessing the page and 
 *      applet as a "file" (vs. "http") url.
 *
 * <LI> Accessing the local file system is (i.e., not using URLs at all)
 *      is problematic because no browser will currently let you do this.
 * </UL>
 * <P>
 * URLFinder will therefore attempt to find a URL by trying the following:
 * <OL>
 * <LI> If the location can be interpreted as an absolute URL (i.e., if it
 *      starts with a protocol) then that URL is used to load the codebook
 *      table.
 *
 * <LI> An attempt is made to load the data using getResource() without
 *      modifying the location string.
 *
 * <LI> If the location does not start with "/", then a "/" is prepended
 *      to the location string and a second attempt is made to use
 *      getResource().
 *
 * <LI> Next the location string is appended to document base and used to
 *      construct a URL.
 *
 * <LI> If the location string includes a path, the document is stripped
 *      off and appended to the document base.
 *
 * <LI> If a codebase is present, the previous two steps are then tried with
 *      the codebase instead of the document base.
 * </OL>
 */
public class URLFinder {
    private static URLFinder instance = null;
    private Hashtable cache = null;

    private URLFinder() {
	cache = new Hashtable();
    }

    /**
     * Returns a URLFinder instance.
     */
    public static URLFinder getURLFinder() {
	if (instance == null)
	    instance = new URLFinder();
	return instance;
    }

    /**
     * Verifies that a URL is, in fact, reachable. This is necessary because
     * the URL constructor only checks for correct URL syntax.
     */
    private URL checkURL(URL url) {
	if (url != null) {
	    try {
	        InputStream tmp = url.openStream();
	        tmp.close();
		url = new URL(url.toString());
	    }
	    catch(Exception e) { 
	        url = null;
	    }
	}
	return url;
    }

    private URL getConcatenatedURL(String base, String location) {
	URL url = null;
	if (base.endsWith("/"))
	    base = base.substring(0, base.length() - 1);
	if (location.startsWith("/"))
	    location = location.substring(1);
	try {
	    url = checkURL(new URL(base + "/" + location));
	}
	catch(Exception e) {
	    url = null;
	}
	return url;
    }

    /**
     * Tries to get a URL by appending the location string to a base url,
     * such as a codebase or document base.
     */
    private URL getURLFromBase(String base, String location) {
	URL url = null;

	// First try just appending them.
	url = getConcatenatedURL(base,location);

	// Then try letting URL append them.
	if (url == null) {
	    try {
	        url = checkURL(new URL(new URL(base), location));
	    }
	    catch(Exception e) {
	        url = null;
	    }
	}
	// Then try just appending the file name
	if (url == null) {
	    try {
	        File f = new File(location);
	        if (f != null)
	            url = getConcatenatedURL(base, f.getName());
	    }
	    catch(Exception e) {
	        url = null;
	    }
	}
	// Then try just appending the file name to the base using the
	// URL appending thingy
	if (url == null) {
	    try {
	        File f = new File(location);
	        if (f != null)
	            url = checkURL(new URL(new URL(base), f.getName()));
	    }
	    catch(Exception e) {
	        url = null;
	    }
	}
	return url;
    }

    /**
     * Attempts to retrieve a URL based on a location string, a class
     * from the jar file containing the data (if the data is stored in 
     * a jar file), an applet codebase, and an applet document base.
     * Other getURL routines provide shortcuts for calling this one.
     */
    public URL getURL(String location, Class classFromJar, String codebase, 
	String documentbase) {

	if (location == null)
	    return null;

	URL url = (URL)cache.get(location);

	// First see if this is a "real" URL
	if (url == null) {
	    try {
	        url = checkURL(new URL(location));
	    }
	    catch(Exception e) {
	        url = null; // Just to make sure...
	    }
	}
	// Next see if we can use getResource
	if (url == null) {
	    if (classFromJar == null)
		classFromJar = getClass();
	    url = checkURL(classFromJar.getResource(location));
	}
	// Next try it without a leading "/". Roumor has it that some
	// browsers don't like the leading "/".
	if (url == null) {
	    String tmpLocation = location;
	    if (tmpLocation.startsWith("/")) {
		tmpLocation = tmpLocation.substring(1);
	        url = checkURL(classFromJar.getResource(location));
	    }
	}

	// See if it's in (or relative to) the document base
	if ((url == null) && (documentbase != null))
	    url = checkURL(getURLFromBase(documentbase, location));

	// See if it's in (or relative to) the code base
	if ((url == null) && (codebase != null))
	    url = checkURL(getURLFromBase(codebase, location));

	// Cache the url if we found one
	if (url != null)
	    cache.put(location, url);
	return url;
    }

    /**
     * Attempts to find a URL based on a location string only. This will
     * work if the location is actually a URL or if it is the location
     * of a resource accessible from this URLFinder instance's classloader.
     * This method should be used only as a last resort.
     */
    public URL getURL(String location) {
	return getURL(location, null, null, null);
    }

    /**
     * Attempts to find a URL based on a location string and an applet.
     * This is the prefered method when using this class in an applet.
     */
    public URL getURL(String location, Applet applet) {
	return getURL(location, applet.getClass(), 
	    applet.getDocumentBase().toString(),
	    applet.getCodeBase().toString());
    }

}


