package sieve.sluice.codebook;

import java.net.*;
import java.util.*;
import java.io.*;
import sieve.sluice.*;

// Modified by Christopher Telfer on 4/7/98

/**
 * A CodeBookTable gets constructed with the name (URL or local file) of 
 * the "codebook" file for a single data file.  This file is parsed
 * to get the following information. <UL>
 * <LI> The name of this data
 * <LI> Name of the data file (an ASCII file with rows separated by newlines)
 * <LI> Whether the rows represent fields or records
 * <LI> The character(s) used as delimiters
 * <LI> The string used for missing values
 * <LI> Attribute-value pairs for this data
 * <LI> A list of fields (variables) with information about each, 
 *      in order of how the fields appear in the data file.
 * </UL>
 * The information about each field includes: <UL>
 * <LI> Its short name (required)
 * <LI> Kind of variable (nominal, ordinal, interval, ratio)
 * <LI> Type that should be used to store values (required).
 * <LI> Its description--a longer string
 * <LI> Value-label mappings for nominal & ordinal data
 * </UL>
 *
 */
public abstract class CodeBookTable extends Table {
    protected URL codebookURL = null;
    protected String dataFile = null;
    protected DataFile data = null;
    protected Hashtable attributes = null;

    protected Variable[] fields = null;
    protected Hashtable fieldIndices = null;    

    protected boolean rowsAreFields;
    protected String dataDelim; 
    protected String missingVal;

    protected int numRecords;


    /**
     * Returns the appropriate subclass of CodeBookTable based on the
     * file extension.
     */
    public static CodeBookTable getTable(URL codebookURL)
	throws InterchangeException {
	if (codebookURL == null)
	    throw new InterchangeException("Null codebook URL");
	String codebook = codebookURL.toString();
	if (codebook.endsWith(".dbc")) return new SGMLTable(codebookURL);
	else if (codebook.endsWith(".sps")) return new SPSSTable(codebookURL);
	else if (codebook.endsWith(".cod")) 
	    return new CodataTable(codebookURL);	
        else if (codebook.endsWith(".stf")) return new STFTable(codebookURL); 
        else throw new InterchangeException(
	    "Unrecognized file type for codebook");
    }
	    
    /**
     * Returns the appropriate subclass of CodeBookTable based on the
     * file extension.
     */
    public static CodeBookTable getTable(String codebook)
	throws InterchangeException {
	URL url = URLFinder.getURLFinder().getURL(codebook);
	return getTable(url);
    }

    public CodeBookTable() throws InterchangeException {}
        
    public CodeBookTable(String codebook) throws InterchangeException {
        setLocation(codebook);
    }

    public CodeBookTable(URL codebook) throws InterchangeException {
        setLocationURL(codebook);
    }
    
    public String getTableType() {
        return "Codebook Data";
    }

    public URL getLocationURL() {
       return codebookURL;
    }

    public void setLocationURL(URL codebookURL) throws InterchangeException {
        dataFile = null;
        attributes = fieldIndices = null;
        fields = null;
        data = null;

        try {
            // Open codebook file & parse, this sets DataFile data also
	    this.codebookURL = codebookURL;
            System.out.println("Parsing the codebook file.");
            parseCodebookFile();    // Sets most of the global variables
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new InterchangeException
                ("Error reading from codebook file " + codebookURL);
        }
        try {
            // Now create DataFile object (find newlines),
            // and figure out numRecords
            System.out.println("Parsing the data file.");
            
//            data = codebook.startsWith("http://")
//                   ? new DataFile(codebookFile, dataFile)
//                   : new RADataFile(codebookFile, dataFile);
            // We had thought that using a RandomAccessFile would be faster
            // for local files.  But it's not because it's not buffered.
            data = new DataFile(codebookURL, dataFile);
            
            if (! rowsAreFields) numRecords = data.lineCount();
            else {   // Have to count tokens in a line--assumes no extra spaces
                data.open();
                StringTokenizer linetok =
                    new StringTokenizer(data.readLine(), dataDelim);
                numRecords = linetok.countTokens();
                data.close();
            }
	    eventHandler.fireTableViewEvent(new TableViewEvent(this));
        }
        catch(Exception e) { 
           e.printStackTrace();
           throw new InterchangeException
               ("Error reading from data file " + dataFile);
        }

    }

    public void insert(Object[][] data, int atRow)
        throws InterchangeException {
        throw new ReadOnlyException();
    }

    public void append(Object[][] data) throws InterchangeException {
        throw new ReadOnlyException();
    }

    public void addField(FieldDescriptor fd) throws InterchangeException {
        throw new ReadOnlyException();
    }

    public void remove(int firstRecord, int lastRecord) throws 
        InterchangeException {
        throw new ReadOnlyException();
    }

    public void removeField(String fieldName) throws InterchangeException {
        throw new ReadOnlyException();
    }

    /* -------------  TableView methods ------------- */
    public TableView table() {
        return this;
    }

    public TableDescriptor getTableDescriptor() {
        TableDescriptor temp = new TableDescriptor();
        for (Enumeration e = attributes.keys(); e.hasMoreElements(); ) {
            Object key = e.nextElement();
            if (key instanceof String)
                temp.setAttribute((String)key, attributes.get(key));
        }
        return temp;
    }

    public String[] getFieldNames() {
	if (fields == null)
	    return new String[0];
	else {
            int n = fields.length;
            String[] flds = new String[n];
            for (int i = 0; i < n; i++)
                flds[i] = fields[i].getName();
            return flds;
	}
    }

    /**
     * Returns the FieldDescriptor for a specific field.
     * WARNING: This gives the user a pointer to our internal storage
     * information about a Field, which can be used to modify it.
     * Not sure if this is good or bad.
     */
    public FieldDescriptor getFieldDescriptor(String field) throws 
        InterchangeException {
        // This will throw an exception if field doesn't exist
        Variable fld = fields[fieldIndex(field)];

        // Temporary implementation of setting the level--PLI
        // Only do if level hasn't already been set.
        if (fld.getLevel() == FieldDescriptor.LEVEL_LABEL) {
            if (fld.getLabels() != null)
              fld.setLevel(FieldDescriptor.LEVEL_NOMINAL);
            else
              fld.setLevel(FieldDescriptor.LEVEL_INTERVAL);
        }
        return fld; 
    }

    public int getRecordCount() { 
	if (fields == null)
	    return 0;
	else
            return numRecords; 
    }

    public Object[][] get(String[] flds, int startingAt,
        int numberOfRecords) throws InterchangeException {
        Object[][]  result  = null;

	if (fields == null) 
	    throw new InterchangeException("No source URL");

        try {
            data.open();

            StringTokenizer linetok = null; 
            result = new Object[flds.length][numberOfRecords];
            int[] map = new int[fields.length];

            // map the "real" field indices to positions in the result
            // array. Need to check for invalid field names.
            for (int i = 0; i < fields.length; i++)  map[i] = -1;
            for (int i = 0; i < flds.length; i++) 
                map[fieldIndex(flds[i])] = i;

            int fldsRequested = flds.length;
            int fldsFound = 0;
            Object value    = null;

            if (rowsAreFields) {
                for (int fld = 0;
                    (fld < fields.length) && (fldsFound < fldsRequested);
                    fld++) {
                    
                    // See if this is a field they want--if not move on
                    if (map[fld] >= 0) {
                        fldsFound++;
                        String fieldType = fields[fld].getType();
                        
                        // Skip in data file to correct place
                        data.skipToLine(fld);
                        // Read the line & update the file location
                        String line = data.readLine();
                        linetok = new StringTokenizer(line, dataDelim);
                        
                        // Skip all of the records up to startingAt
                        for (int rec = 0; rec < startingAt; rec++) 
                            linetok.nextToken();

                        // Copy the requested number of records into the array
                        for (int rec = 0; rec < numberOfRecords; rec++) {
                            String  cellString  = linetok.nextToken();
                            value = null;
                            if (! cellString.equals(missingVal)) {
                                value = classToType(cellString, fieldType);
                            }
                            result[map[fld]][rec] = value;
                        }
                    }
                }
            }
            else {  // Rows are records
                data.skipToLine(startingAt);

                String line;
                int rec;
                // Iterate over each record (row)
                for (rec = 0, line = data.readLine();
                    (rec < numberOfRecords) && 
                    (line != null); rec++, line = data.readLine()) {
                    
                    linetok = new StringTokenizer(line, dataDelim);

                    // Within each row, iterate over each field
                    for (int fld = 0, fldcount = 0; (fld < fields.length) && 
                        linetok.hasMoreTokens(); fld++) {
                        if (fldcount == flds.length) break; // optimize

                        String cellString = linetok.nextToken();
                        // If this field wasn't requested by the user, no
                        // further work must be done.
                        if (map[fld] >= 0) {
                            value = null;
                            if (! cellString.equals(missingVal)) {
                                value = classToType(cellString,
                                    fields[fld].getType());
                            }
                            result[map[fld]][rec] = value;
                                        fldcount++;
                        }
                    }
                }
            }
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new InterchangeException(e.toString());
        }
        finally {
            data.close();
        }
        
        return result;
    }

    public Object[][] get(String[] flds, int startingAt)
        throws InterchangeException {
        return get(flds, startingAt, numRecords - startingAt);
    }

    public Object[] get(String fldname) throws InterchangeException {
        String[] flds = { fldname };
        return (get(flds, 0, numRecords))[0];
    }

    /* ------------------ Private methods -------------------------- */

    /**
     * Parses the codebook file for all of the required information.
     * Sets the following class variables:<UL>
     * <LI> dataFile
     * <LI> rowsAreFields
     * <LI> dataDelim
     * <LI> missingVal
     * <LI> attributes
     * <LI> fields
     * <LI> fieldIndices
     * </UL>
     */
    protected abstract void parseCodebookFile() throws Exception;

        
    /**
     * Method to turn a string into the Java type specified by another
     * string.
     */
    private static Object classToType(String s, String storedType) 
        throws NumberFormatException {
        Object returnObj = null;
        char hashVal = storedType.charAt(10);
        switch (hashVal) {
          case 'I': returnObj = new Integer(s); break;
          case 'L': returnObj = new Long(s); break;
          case 'F': returnObj = new Float(s); break;
          case 'D': returnObj = new Double(s); break;
          case 'S': returnObj = s; break;
        }
        return returnObj;
    }

    /**
     * Adds an attribute to this table's list of attributes
     */
    protected void setAttribute(String name, String value) {
        if (attributes == null) attributes = new Hashtable();
        attributes.put(name, value);
    }
    
    /**
     * Return the mapping from a field name to its index.
     * Throws an exception if no field exists with that name
     */
    private int fieldIndex(String fieldName) throws InterchangeException {
        Object intObj = fieldIndices.get(fieldName);
        if (intObj == null)
            throw new sieve.sluice.NoSuchFieldException(fieldName);
        return ((Integer)intObj).intValue();
    }


}    









