package sieve.sluice.codebook;

import java.net.*;
import java.util.*;
import java.io.*;
import sieve.sluice.*;

public class SGMLTable extends CodeBookTable {
    
    public SGMLTable(String url) throws InterchangeException {
        super(url);
    }

    public SGMLTable(URL url) throws InterchangeException {
        super(url);
    }
    
    public SGMLTable() throws InterchangeException {
        //super("http://simon.cs.vt.edu/~jnielsen/data/mean.dbc");
    }

    /**
     * Parses the codebook file for all of the required information.
     * Sets the following class variables:<UL>
     * <LI> dataFile
     * <LI> rowsAreFields
     * <LI> dataDelim
     * <LI> missingVal
     * <LI> attributes
     * <LI> fields
     * <LI> fieldIndices
     * </UL>
     */
    protected void parseCodebookFile() throws IOException, 
        MalformedURLException,
        InterchangeException, SyntaxException {

	URL book = getLocationURL();
	if (book == null)
	    return;
        InputStream is = book.openStream();

        LineNumberInputStream cbStream = 
            new LineNumberInputStream(new BufferedInputStream(is));

        // Put "name" from DATABASE tag 
        Tag currTag = SGMLParser.getNextTag(cbStream);
        if (! currTag.name.equals("DATABASE"))
            throw new SyntaxException(cbStream.getLineNumber());
        // JMN: What do we do with this name??? 

        // Get dataFile name from DATFILE tag
        currTag = SGMLParser.getNextTag(cbStream);
        if (! currTag.name.equals("DATAFILE"))
            throw new SyntaxException(cbStream.getLineNumber());

        // Create string for data file in same context as codebook
        dataFile = currTag.attribute("NAME");

        // Set rowsAreFields, dataDelim, and missingVal from DATAFILE tag
        rowsAreFields = (currTag.attribute("ROW").equalsIgnoreCase("RECORD"))
            ? false : true;     
        dataDelim = currTag.attribute("DELIM");
        missingVal = currTag.attribute("MISSING");

        // Put attributes from ATTR tags into "attributes" Hashtable
        for (currTag = SGMLParser.getNextTag(cbStream);
            currTag.name.equals("ATTR");
            currTag = SGMLParser.getNextTag(cbStream)) {
            setAttribute(currTag.attribute("NAME"), currTag.attribute("VAL"));
        }
    
        // Now read variable information from remaining tags
        // Set fields and fieldIndices
        VariableVector tempFields = new VariableVector();
        fieldIndices = new Hashtable();
        int varCount = 0;
        Variable currVar = null;
        while(! currTag.name.equals("EOF")) {    // Loop until EOF
            if (currTag.name.equals("VARIABLE")) {   // VARIABLE tag
                String varType = pickType(currTag.attribute("TYPE"));
                currVar = new Variable(currTag.attribute("NAME"), varType);
                tempFields.addElement(currVar);
                fieldIndices.put(currVar.getName(), new Integer(varCount++));
                // Now set level if provided
                String level = currTag.attribute("LEVEL");
                if (level != null) currVar.setLevel(stringToLevel(level));
            }

            else if (currTag.name.equals("LABEL")) {   // LABEL tag
                if (currVar == null) 
                    throw new SyntaxException(cbStream.getLineNumber());
                currVar.setDescription(currTag.attribute("STRING"));
            }

            else if (currTag.name.equals("VALLABEL")) {  // VALLABEL tag
                if (currVar == null) 
                    throw new SyntaxException(cbStream.getLineNumber());
                Integer code = (new Integer(currTag.attribute("CODE")));
                currVar.addValLabel(code, currTag.attribute("MEANING"));
            }

            else if (currTag.name.equals("ATTR")) {    // ATTR tag
                if (currVar == null)
                    throw new SyntaxException(cbStream.getLineNumber());
                String attrName = (String)currTag.attribute("NAME");
                currVar.addAttr(attrName.toUpperCase(),
                    currTag.attribute("VALUE"));
            }
            // Ignore any other tags
            currTag = SGMLParser.getNextTag(cbStream);
        }
        fields = tempFields.getArray();

        // Close codebook file
        cbStream = null;
    }

    /**
     * Chooses a storage type for a variable.  Right now it just
     * reads the declared type & maps it to a Java type.  We may want to
     * make this more intelligent in the future--depending on the
     * specified range, etc.
     */
    protected static String pickType(String declaredType) {
        String retClass = null;
        try {
            if (declaredType.equalsIgnoreCase("INT")) 
                retClass = "java.lang.Integer";
            else if (declaredType.equalsIgnoreCase("LONG")) 
                retClass = "java.lang.Long";
            else if (declaredType.equalsIgnoreCase("STRING")) 
                retClass = "java.lang.String";
            else if (declaredType.equalsIgnoreCase("DEC")) 
                retClass = "java.lang.Float";
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        if (retClass == null)
            System.err.println(
                "WARNING: No class found for type: " + declaredType);
        return retClass;
    }

    protected static int stringToLevel(String s) {
        int retVal = FieldDescriptor.LEVEL_LABEL;
        switch(Character.toUpperCase(s.charAt(0))) {
         case 'N':
           retVal = FieldDescriptor.LEVEL_NOMINAL; break;
         case 'O':
           retVal = FieldDescriptor.LEVEL_ORDINAL; break;
         case 'I':
           retVal = FieldDescriptor.LEVEL_INTERVAL; break;
         case 'R':
           retVal = FieldDescriptor.LEVEL_RATIO; break;
        }
        return retVal;
    }
        
        

}    









