package sieve.sluice.views;
 
import sieve.sluice.*;
import sieve.sluice.event.*;
import java.util.*;
import java.beans.*;
import sieve.sluice.views.predicates.*;

/**
 * A CriteriaView filters the records presented by its source and
 * generates a view of a table containing only those records which
 * meet a specified criteria.
 * Criteria are specified by unary or binary predicates.
 * <P>
 * Classes which actually perform tests are currently implemented in the
 * <CODE>sieve.sluice.views.predicates</CODE> package. Currently supported
 * tests include:
 * <P>
 * For all Objects: "isEqualTo", "isMissing"
 * <BR>
 * For Strings: "isEqualTo", "isEqualToIgnoringCase", 
 * "isLexicographicallyBefore", "isLexicographicallyAfter",  "endsWith", 
 * "startsWith", "contains", and "containsIgnoringCase"
 * <BR>
 * For Numbers: "isEqualTo", "isLessThan", "isLessThanOrEqualTo", 
 * "isGreaterThan", and "isGreaterThanOrEqualTo"
 * <BR>
 * For Dates: "isEqualTo", "isBefore", "isAfter", 
 * <P>
 * The methods
 * <CODE>sieve.sluice.views.predicates.Predicate.getBinaryPredicateNames(Class)</CODE>
 * and
 * <CODE>sieve.sluice.views.predicates.Predicate.getUnaryPredicateNames(Class)</CODE>
 * may be used to get a list of predicates for an abitrary class. Test names
 * may include spaces between words.
 * <P>
 * As an example, consider a <CODE>PhoneBookView</CODE> class which provided
 * address information. To get a table of all of the Smiths, you could use:
 * <PRE>
 *  CriteriaView smithView = new CriteriaView();
 *  smithView.addSourceView(phoneBookView);
 *  smithView.setField("Last Name");
 *  smithView.setPredicate("isEqualTo");
 *  smithView.setValue("Smith");
 * </PRE>
 */
public class CriteriaView extends SampleView implements Sampler {

    protected FieldDescriptor 	fd 	= null;
    protected Vector		matches	= null;
    protected Object		value	= null;
    protected int		relation = 0;
    protected boolean		testAsString = false;
    protected boolean		negatePredicate = false;
    protected Predicate		predicate = null;
    protected String   		predicateName = null;

    /**
     * Construct a new CriteriaView with the given source.
     */
    public CriteriaView(TableView source) {
	super(source, null);
	setSampler(this);
    }

    /**
     * Construct a new CriteriaView. The predicate defaults to "Is Equal 
     * To".
     */
    public CriteriaView() {
	super(null, null);
	setSampler(this);
	setPredicate("IsEqualTo"); // We always have this one...
    }

    /**
     * Returns the field to which the test will be applied. Returns null
     * if no field has been set.
     */
    public String getField() {
	if (fd == null)
	    return null;
	else
	    return fd.getName();
    }

    /**
     * Sets the field to which the test will be applied.
     */
    public void setField(String field) throws InterchangeException {
	if ((fd == null) || (! field.equals(fd.getName()))) {
	    String oldField = getField();
	    fd = source.getFieldDescriptor(field);
	    setupPredicate();
	    reset();
	    changes.firePropertyChange("field", oldField, fd.getName());
	}
    }

    public synchronized void setValue(Object value) {
	if ((this.value == null) || (value == null) || 
	    (! value.equals(this.value))) {
	    Object oldValue = this.value;
	    this.value = value;
	    reset();
	    changes.firePropertyChange("value", oldValue, value);
	}
    }

    public Object getValue() {
	return value;
    }

    private synchronized void setupPredicate() {
	predicate = null;
	if (predicateName != null) {
	    if (testAsString)
	        predicate = Predicate.getPredicateFromName(String.class, 
		    predicateName);
	    else if (fd != null) {
	        try {
	            predicate = Predicate.getPredicateFromName(
		        Class.forName(fd.getType()), predicateName);
	        }
	        catch (Exception e) {
		    e.printStackTrace();
	        }
	    }
	}
    }

    // Need to figure out how to make this more efficient...
    public synchronized void setPredicate(String predicateName) {
/* PROBLEM: 
	if ((predicateName == null) || (this.predicateName == null) 
	    || (! predicateName.equals(this.predicateName))
	    ) {
	if (predicateName != null) {
*/
	    String oldPredicateName = this.predicateName;
	    this.predicateName = predicateName;
	    setupPredicate();
	    reset();
	    changes.firePropertyChange("predicate", oldPredicateName, 
		predicateName);
/*
	}
*/
    }

    public String getPredicate() {
	return predicateName;
    }

    public boolean getTestAsString() {
	return testAsString;
    }

    public synchronized void setTestAsString(boolean testAsString) {
	if (testAsString != this.testAsString) {
	    this.testAsString = testAsString;
	    setupPredicate();
	    reset();
	    changes.firePropertyChange("testAsString", 
		new Boolean(! testAsString), new Boolean(testAsString));
	}
    }

    public boolean getNegatePredicate() {
	return negatePredicate;
    }

    public synchronized void setNegatePredicate(boolean negatePredicate) {
	if (negatePredicate != this.negatePredicate) {
	    this.negatePredicate = negatePredicate;
	    reset();
	    changes.firePropertyChange("negatePredicate", 
		new Boolean(! negatePredicate), new Boolean(negatePredicate));
	}
    }

    public void addSourceView(TableView source) throws TooManySourcesException {
	if (this.source != null)
            throw new TooManySourcesException("Only one source allowed.");
	this.source = source;
	eventHandler.fireTableViewEvent(new SourceAddedEvent(this));
	try { if (source.getFieldNames().length >0 )
	    setField(source.getFieldNames()[0]);
        }
        catch (TooManySourcesException e) {
	    throw e;
	}
        catch (InterchangeException ex) {
	    // WHAT TO DO HERE?
      	}
    }

    private void reset() {
//System.out.println("In Crit.reset() with val = " + value + " pred = " + predicate + " fd = " + fd);
	try {
	    if ((value != null) && (fd != null) && (predicate != null)) {
		Object first = null;
		Object second = value;	
		String[] fldName = {fd.getName()};
		if (testAsString) 
		    second = value.toString();

		matches = new Vector();
		Object[][] fld = source.get(fldName, 0);
		if ((value == null) && (fld.length > 0) && (fld[0].length > 0))
		    value = fld[0][0];
		boolean add = false;
		for (int i=0; i < fld[0].length; i++) {
		    //if (fld[0][i] != null) { // Handle missing values
		        if (testAsString && (fld[0][i] != null))
			    first = fld[0][i].toString();
			else 
			    first = fld[0][i];	
			
			if (predicate instanceof UnaryPredicate)
			    add = ((UnaryPredicate)predicate).execute(first);
			else
			    add = ((BinaryPredicate)predicate).execute(first,
				second);
		        if ((add && ! negatePredicate) || 
			    (!add && negatePredicate)) {
			    matches.addElement(new Long(i));
			}
		    //}
		}
	        eventHandler.fireTableViewEvent(new TableViewEvent(this));
	    }
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }

    public synchronized int getRecordNumber(int requestedRecord) throws
        NoSuchRecordException {
	if ((source == null) || (requestedRecord >= matches.size()))
		throw new NoSuchRecordException(requestedRecord);
	return ((Long)matches.elementAt((int)requestedRecord)).intValue();
    }

    public synchronized int size() {
	if ((source == null) || (matches == null))
	    return 0;
	else
	    return matches.size();
    }

    public synchronized boolean recordIsInSample(int record) {
	if ((source == null) || (matches == null))
	    return false;
	else
	    return matches.contains(new Long(record));
    }

    public synchronized void tableViewChanged(TableViewEvent event) {
        reset();
        eventHandler.fireTableViewEvent(event);
    }

}
