package sieve.sluice.views;

import sieve.sluice.*;
import sieve.sluice.event.*;
import java.util.*;
import java.beans.*;

/**
 * The default view of a table is just a wrapper whose methods call the
 * corresponding methods on the view from which we were instantiated.
 * The intent here is that this class <I>can</I> serve as a superclass
 * for other <CODE>TableView</CODE>s.
 */
public class DefaultTableView implements TableView  {
    TableView			source		= null;
    TableViewEventSupport	eventHandler	= null;
    PropertyChangeSupport 	changes         = null;

    public DefaultTableView(TableView source) {
	this.source = source;
	this.eventHandler = new TableViewEventSupport();
        this.changes = new PropertyChangeSupport(this);
    }

    public DefaultTableView() {
	this(null);
    }

    public TableView table() throws NoSourceException {
	if (source == null)
	    throw new NoSourceException();
	return source.table();
    }

    public TableDescriptor getTableDescriptor() throws NoSourceException {
	if (source == null)
	    throw new NoSourceException();
	return source.getTableDescriptor();
    }

    public FieldDescriptor getFieldDescriptor(String field) throws 
        InterchangeException {
	if (source == null)
	    throw new NoSourceException();
	return source.getFieldDescriptor(field);
    }

    public String[] getFieldNames() throws NoSourceException {
	if (source == null)
	    throw new NoSourceException();
	return source.getFieldNames();
    }

    /**
     * Returns number of records in the view.
     */
    public int getRecordCount() throws NoSourceException {
	if (source == null)
	    throw new NoSourceException();
	return source.getRecordCount();
    }

    public Object[][] get(String[] fields, int startingAt, 
	int numberOfRecords) throws InterchangeException {
	if (source == null)
	    throw new NoSourceException();
	return source.get(fields, startingAt, numberOfRecords);
    }

    /**
     * Returns all records starting from record number <CODE>startingAt</CODE>
     * from the fields specified in <CODE>fields</CODE>.
     */
    public Object[][] get(String[] fields, int startingAt) 
	throws InterchangeException { 
	if (source == null)
	    throw new NoSourceException();
	return source.get(fields, startingAt); 
    } 

    /* -------------  Event-related methods ------------- */
    public void addTableViewListener(TableViewListener listener) {
	eventHandler.addTableViewListener(listener);
    }

    public void removeTableViewListener(TableViewListener listener) {
	eventHandler.removeTableViewListener(listener);
    }

    /**
     * Propagates event to all listeners with no filtering.
     */
    public void tableViewChanged(TableViewEvent event) {
	eventHandler.fireTableViewEvent(event);
    }

    /* -------------  Methods for link support ------------- */
    /**
     * By default we throw an exception if we already have a source.
     * If we don't, then an event is sent to all of our listeners.
     * We DO NOT add ourselves as a listener to the new source. This
     * is the responsibility of whatever is connecting us.
     */
    public void addSourceView(TableView source) throws TooManySourcesException {
        if (this.source != null)
	    throw new TooManySourcesException("Only one source allowed.");
	this.source = source;
	eventHandler.fireTableViewEvent(new SourceAddedEvent(this));
    }

    /**
     * If the parameter is actually our current source, we remove it
     * and fire an event to anyone listening to us.
     */
    public void removeSourceView(TableView source) {
        if (source == this.source) {
            this.source = null;
	    eventHandler.fireTableViewEvent(new SourceRemovedEvent(this));
	}
    }

    //----------------------------------------------------------------------
    // Methods for registering listeners:
    /**
     * The specified PropertyChangeListeners <b>propertyChange</b> method will
     * be called each time the value of any bound property is changed.
     *
     * @see #removePropertyChangeListener
     * @param l the PropertyChangeListener
     */
    public void addPropertyChangeListener(PropertyChangeListener l) {
        changes.addPropertyChangeListener(l);
    }
 
    /**
     * Remove this PropertyChangeListener from the internal list.
     * If the PropertyChangeListener isn't on the list, silently do nothing.
     *
     * @see #addPropertyChangeListener
     * @param l the PropertyChangeListener
     */
    public void removePropertyChangeListener(PropertyChangeListener l) {
        changes.removePropertyChangeListener(l);
    }

}

