package sieve.sluice.views;

import java.util.*;
import sieve.sluice.*;
import sieve.sluice.event.*;

/**
 * A FunctionView generates a view of data generated from a function.
 * The independent variable is generated from a field of a source
 * table. (A subclass of this should allow generation of a function
 * with explicit min and max values.)
 */
public class FunctionView extends DefaultTableView {
    protected int numberOfPoints	= 0;
    protected String indepField		= null;

    protected Hashtable fieldDescriptors		= null;
    protected Vector functions = null;
    
    public FunctionView() {
    }

    public void addSourceView(TableView source) throws TooManySourcesException {
        if (this.source != null)
            throw new TooManySourcesException("Only one source allowed.");
        this.source = source;

	NumberView nv = new NumberView(source);
	indepField = null;
	try {
	    TableDescriptor td = nv.getTableDescriptor();
	    indepField = td.getIndependentVariableField();
	    if (indepField != null) {
	        FieldDescriptor fd = nv.getFieldDescriptor(indepField);
	    }
	}
  	catch (InterchangeException e) {
	    // The table descriptor had a bogus indep var field name
	    indepField = null;
	}

	// If the table descriptor either didn't specify an indep var
	// or if the specified var wasn't present or numeric, try to
	// use the first (numeric) field.
	if (indepField == null) {
	    try {
	        String[] flds = nv.getFieldNames();
	        if (flds.length > 0)
		    indepField = flds[0];
	    }
	    catch(NoSourceException ex) {}
	}
	try {
	    setNumberOfPoints(nv.getRecordCount());
	}
	catch (InterchangeException e) {
	    setNumberOfPoints(0);
	}
	functions = new Vector();
	fieldDescriptors = new Hashtable();
	
	reset();

        eventHandler.fireTableViewEvent(new SourceAddedEvent(this));
    }

    public void removeSourceView(TableView source) {
	indepField = null;
	functions = null;
	fieldDescriptors = null;
	super.removeSourceView(source);
    }

    public String getIndependentField() {
	return indepField;
    }
    public void setIndependentField(String indepField) throws 
  	InterchangeException {
	if (source == null)
	    throw new NoSourceException();
	NumberView nv = new NumberView(source);
	// See if this bombs
	FieldDescriptor fd = nv.getFieldDescriptor(indepField);
	if (fd == null) // Shouldn't happen, but just in case
	    throw new sieve.sluice.NoSuchFieldException(indepField);
	this.indepField = indepField;
	reset();
    }

    public void addFunction(Function f) {
	if (functions != null) {
	    functions.addElement(f);
	    fieldDescriptors.put(f.getName(), new FunctionFieldDescriptor(f));
	    resetDependent();
            eventHandler.fireTableViewEvent(new TableViewEvent(this));
	}
    }
    public void removeFunction(Function f) {
	if (functions != null) {
	    functions.removeElement(f);
	    fieldDescriptors.remove(f.getName());
            eventHandler.fireTableViewEvent(new TableViewEvent(this));
	}
    }
    

    protected void reset() {
	resetIndependent();
	resetDependent();
    }

    protected void resetIndependent() {
	if ((fieldDescriptors != null) && (functions != null)) {
	    try {
		NumberView nv = new NumberView(source);

		String[] indepFldArray = {indepField};
		Object[][] indepDataObjs = nv.get(indepFldArray, 0);
		double[] indepData = makeDoubleArray(indepDataObjs[0]);

		// Setup the new independent var values first
		FunctionFieldDescriptor fd = 
		    new FunctionFieldDescriptor(
		    nv.getFieldDescriptor(indepField));
		fieldDescriptors.put(indepField, fd);
		double[] newIndepValues = new double[numberOfPoints];
		double step = ((double)indepData[indepData.length - 1] -
		    (double)indepData[0])/((double)numberOfPoints - 1);

		newIndepValues[0] = indepData[0];
		for (int r = 1; r < numberOfPoints; r++)
		    newIndepValues[r] = newIndepValues[r - 1] + step; 
		fd.setValues(newIndepValues);
	    } 
	    catch(InterchangeException e) {
		e.printStackTrace();
	    }
	}
    }

    protected void resetDependent() {
	if ((fieldDescriptors != null) && (functions != null)) {
	    FunctionFieldDescriptor fd = 
		    (FunctionFieldDescriptor)fieldDescriptors.get(indepField);

	    double[] newIndepValues = fd.getPrimativeValues();

	    // Now do all of the dependent var fields
	    for (int f = 0; f < functions.size(); f++) {
	        Function fn = (Function)functions.elementAt(f);

	        double newValues[] = new double[numberOfPoints];
		    
		for (int r = 0; r < numberOfPoints; r++) {
		    newValues[r] = fn.evaluate(newIndepValues[r]);
		}
		((FunctionFieldDescriptor)fieldDescriptors.get(
		    fn.getName())).setValues(newValues); 
	    }
            eventHandler.fireTableViewEvent(new TableViewEvent(this));
	}
    }

    protected double[] makeDoubleArray(Object[] array) {
	double[] dArray = new double[array.length];
	for (int i = 0; i < array.length; i++) {
	    dArray[i] = ((Number)(array[i])).doubleValue();
	}
	return dArray;
    }

    public int getRecordCount() throws NoSourceException {
	if (source == null)
	    throw new NoSourceException();
	source.getRecordCount(); // Give this a chance to throw an exception
	return getNumberOfPoints();
    }
    
    public int getNumberOfPoints() {
	return numberOfPoints;
    }    
    public void setNumberOfPoints(int numberOfPoints) {
	int oldNumber = this.numberOfPoints;
	try {
	    if (source != null && numberOfPoints < source.getRecordCount())
		numberOfPoints = source.getRecordCount();
	}
	catch (InterchangeException e) { }
	this.numberOfPoints = numberOfPoints;
	reset();
	changes.firePropertyChange("numberOfPoints", new Integer(oldNumber),
	    new Integer(numberOfPoints));
    }

    public Object[][] get(String[] fields, int startingAt,
        int numberOfRecords) throws InterchangeException {
        if (source == null)
            throw new NoSourceException();
	if (startingAt > numberOfPoints)
            throw new NoSuchRecordException(startingAt);
	if (startingAt + numberOfRecords > numberOfPoints)
            throw new NoSuchRecordException(numberOfPoints);

	Object[][] data = new Object[fields.length][];

	for (int f = 0; f < fields.length; f++) {
	    FunctionFieldDescriptor fd = 
		(FunctionFieldDescriptor)fieldDescriptors.get(fields[f]);
	    if (fd == null)
		throw new sieve.sluice.NoSuchFieldException(fields[f]);
	    if (numberOfRecords == numberOfPoints)
		data[f] = fd.getValues();
	    else {
		data[f] = new Object[numberOfRecords];
		Double[] values = fd.getValues();
		for (int r = 0; r < numberOfRecords; r++)
		    data[f][r] = values[r + startingAt];
	    }
	}
	return data;
    }

    public Object[][] get(String[] fields, int startingAt)
        throws InterchangeException {
	return get(fields, startingAt, getRecordCount() - startingAt);
    }

    public String[] getFieldNames() throws NoSourceException {
        if ((source == null) || (functions == null))
            throw new NoSourceException();
	String[] flds = new String[functions.size() + 1];
	flds[0] = indepField;
	for (int i = 1; i < flds.length; i++)
	    flds[i] = ((Function)(functions.elementAt(i - 1))).getName();
        return flds;
    }

    public FieldDescriptor getFieldDescriptor(String field) throws
        InterchangeException {
        if (source == null)
            throw new NoSourceException();
	FunctionFieldDescriptor fd = 
	    (FunctionFieldDescriptor)fieldDescriptors.get(field);
	if (fd == null)
	    throw new sieve.sluice.NoSuchFieldException(field);

	return fd;
    }

    public void tableViewChanged(TableViewEvent event) {
	reset();
        eventHandler.fireTableViewEvent(event);
    }


}

class FunctionFieldDescriptor extends FieldDescriptor {
    private transient Double[] values = null;
    private transient double[] primativeValues = null;
    private Function  f = null;

    public FunctionFieldDescriptor(FieldDescriptor origDescriptor) {
	super(origDescriptor.getName());
	copy(origDescriptor);
	if (origDescriptor instanceof Function)
	    f = (Function)origDescriptor;
    }

    public String getType() {
	return Double.class.getName();
    }

    Double[] getValues() {
	return values;
    }

    double[] getPrimativeValues() {
	if ((primativeValues == null) && (values != null)) {
	    primativeValues = new double[values.length];
	    for (int i = 0; i < values.length; i++)
                primativeValues[i] = values[i].doubleValue();
	}
	return primativeValues;
    }

    void setValues(Double[] values) {
	this.values = values;
	this.primativeValues = null;
    }

    void setValues(double[] vals) {
	this.primativeValues = vals;
	values = new Double[vals.length];
	for (int i = 0; i < vals.length; i++)
	    values[i] = new Double(vals[i]);
    }

}

