package sieve.sluice.views;

import java.util.*;
import sieve.sluice.*;
import sieve.sluice.event.*;

/**
 * An PolyInterpolatedView generates a table of arbitrary size containing
 * values interpolated from numeric fields.
 * One (numeric) field is assumed to contain the independent variable
 * and all other numeric fields are assumed to be dependent variables.
 * The values in the independent variable field must be ordered.
 */
public class PolyInterpolatedView extends FunctionView {
    protected int orderOfFit		= 0;
    protected Hashtable depFields	= null;

    public PolyInterpolatedView() {

    }

    public void addSourceView(TableView source) throws TooManySourcesException {
	super.addSourceView(source);
	depFields = new Hashtable();

	resetFunctions();
    }

    public void removeSourceView(TableView source) {
	depFields = null;
	super.removeSourceView(source);
    }

    public void setIndependentField(String indepField) throws 
        InterchangeException {
	super.setIndependentField(indepField);
	depFields = new Hashtable();
	resetFunctions();
    }

    protected void resetFunctions() {
	if (depFields != null) {
	    try {
		NumberView nv = new NumberView(source);
		String[] flds = nv.getFieldNames();

		String[] indepFldArray = {indepField};
		Object[][] indepDataObjs = nv.get(indepFldArray, 0);
		double[] indepData = makeDoubleArray(indepDataObjs[0]);
		Object[][] allData = nv.get(flds, 0);

		// Make a function for of the dependent var fields
		for (int f = 0; f < flds.length; f++) {
		    if (! flds[f].equals(indepField)) {

		        PolyInterpolatedFunction fn = 
			    new PolyInterpolatedFunction (
			    nv.getFieldDescriptor(flds[f]));
		        depFields.put(flds[f], fn);

			double depValues[] = makeDoubleArray(allData[f]);
		    
			// Coefficients of the polynomial
			double c[] = findCoefficients(indepData, depValues);
			fn.setConstants(c);
			super.addFunction(fn);
		    }

		}
        	eventHandler.fireTableViewEvent(new TableViewEvent(this));
	    }
	    catch(InterchangeException e) {
		e.printStackTrace();
	    }

	}
	
    }

    // Shamelessly adapted from PolyFit
    private double[] findCoefficients(double[] indepVals, double[] depVals) {
	int rows = indepVals.length;
	int cols = depVals.length;
        int order = orderOfFit + 1;

	double basis[] = new double[order * rows];
	double alpha[] = new double[order * order];
	double alpha2[] = new double[order * order];
	double beta[] = new double[order];

	// calc basis
        for (int i = 0; i < order; i++) {
            for (int j = 0; j < rows; j++) {
                int k = i + j * order;
 
                if  (i == 0)
                    basis[k] = 1;
                else
                    basis[k] = indepVals[j] * basis[k - 1];
            }
	}

	// calc alpha2
        for (int i = 0; i < order; i++) {
            for (int j = 0; j <= i; j++) {
                double sum = 0;
                for (int k = 0; k < rows; k++) {
                    int k2 = i + k * order;
                    int k3 = j + k * order;
 
                    sum += basis[k2] * basis[k3];
                }
 
                int k2 = i + j * order;
                alpha2[k2] = sum;
 
                if (i != j) {
                    k2 = j + i * order;
                    alpha2[k2] = sum;
                }
            }
	}

	// calc beta
 
        for (int j = 0; j < order; j++) {
            double sum = 0;
            for (int k = 0; k < rows; k++) {
                int k3 = j + k * order;
 
                sum += depVals[k] * basis[k3];
            }
            beta[j] = sum;
        }
 
        // get alpha
        for (int j = 0; j < order * order; j++)
           alpha[j] = alpha2[j];
 
        //solve for params 
        if (!solve(alpha, beta, order)) {
	    System.err.println("Warning: Matrix solve failed");
        }
	return beta;
    }

    boolean solve(double a[], double b[], int n) {
       for (int i = 0; i < n; i++) {
           //find pivot
           double mag = 0;
           int pivot = -1;
 
           for (int j = i; j < n; j++) {
           	double mag2 = Math.abs(a[i + j * n]);
           	if  (mag2 > mag) {
                    mag = mag2;
                    pivot = j;
           	}
       	    }
 
            //no pivot:error
       	    if  (pivot == -1 || mag == 0) {
           	//app->message_alert("Solver error: singular matrix");
           	return false;
       	    }
       	    //move pivot row into position
       	    if (pivot != i) {
           	double temp;
           	for (int j = i; j < n; j++) {
               	    temp = a[j + i * n];
               	    a[j + i * n] = a[j + pivot * n];
               	    a[j + pivot * n] = temp;
           	}
           	temp = b[i];
           	b[i] = b[pivot];
           	b[pivot] = temp;
       	    }

       	    //normalize pivot row
	    mag = a[i + i * n];
            for (int j = i; j < n; j++)
                a[j + i * n] /= mag;
                b[i] /= mag;
 
                //eliminate pivot row component from other rows
		for (int i2 = 0; i2 < n; i2++) {
		if (i2 == i)
                    continue;
 
		double mag2 = a[i + i2 * n];
		for (int j = i; j < n; j++)
		    a[j + i2 * n] -= mag2 * a[j + i * n];
 
		b[i2] -= mag2 * b[i];
	    }
	}
 
	//result should be identity matrix in a and solution vector in b
   	return true;
    }

    public int getOrderOfFit() {
	return orderOfFit;
    }    
    public void setOrderOfFit(int orderOfFit) {
	int oldNumber = this.orderOfFit;
	if (orderOfFit < 0)
	    return;
	this.orderOfFit = orderOfFit;
	resetFunctions();
	changes.firePropertyChange("orderOfFit", new Integer(oldNumber),
	    new Integer(orderOfFit));
    }

}

class PolyInterpolatedFunction extends Function {
    private transient Double[] values = null;
    private double[] c = null;

    public PolyInterpolatedFunction(FieldDescriptor origDescriptor) {
	super(origDescriptor.getName());
	copy(origDescriptor);
    }

    public void setConstants(double[] c) {
	this.c = c;
    }

    public double evaluate(double x) {
    	double x2 = 1.0;
	double w = 0;
	for (int k = 0; k < c.length; k++) {
	    w += c[k] * x2;
	    x2 *= x;
	}
	return w;
    }

}



