package sieve.whiteboard;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import sieve.workbench.*;
 
/**
 * A Line component. These are created by the LineTool.
 * @see LineTool
 */
public class Line extends Component implements WorkbenchComponent,
	MouseMotionListener, MouseListener {

    static {
	ToolFactory.getToolFactory().registerTool(Line.class, LineTool.class);
	WorkbenchScaler.addScaler(Line.class, new WhiteBoardScaler());
    }

    public static final int HANDLE_SIZE = 3;

    private static int STATE_UNSELECTED = 0;
    private static int STATE_SELECTED   = 1;
    private static int STATE_MOVING     = 2;
    private static int STATE_RESHAPING  = 3;
    int state = STATE_UNSELECTED;
    Point pointBeingMoved = null;

    Point start = null;
    Point end = null;
    Point lastPoint = null;
    Polygon poly = null;
    Workbench wb = null;

    UniqueID id = null;
    private PropertyChangeSupport listeners = null;

    private boolean constructed = false;
    private double length = 0.0;
    private Rectangle startHandle = new Rectangle();
    private Rectangle endHandle = new Rectangle();

    private Color color = Color.black;

    public synchronized UniqueID getUniqueID() {
	return id;
    }

    public synchronized void setUniqueID(UniqueID id) {
	this.id = id;
    }

    public Line(Point start, Point end, Workbench wb) {
	this.start = start;
	this.end = end;
	this.wb = wb;
	this.id = new UniqueID(this);
	addMouseListener(this);
	addMouseMotionListener(this);
	this.listeners = new PropertyChangeSupport(this);
    }

    public Rectangle getBounds() {
	Dimension parentSize = getParent().getSize();
        return new Rectangle(0, 0, parentSize.width, parentSize.height);
    }

    public void setColor(Color color) {
	this.color = color;
    }
    public Color getColor() {
	return color;
    }

    public void paint(Graphics g) {
        g.setColor(color);
        if ((start != null) && (end != null))
            g.drawLine(start.x, start.y, end.x, end.y);

	length = Math.sqrt((start.x - end.x) * (start.x - end.x) +
		(start.y - end.y) * (start.y - end.y));

	if (isSelected()) {
	    startHandle = new Rectangle(start.x - HANDLE_SIZE, 
		start.y - HANDLE_SIZE, 
		2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
	    endHandle = new Rectangle(end.x - HANDLE_SIZE, end.y - HANDLE_SIZE, 
		2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
	   
	    g.fillRect(start.x - HANDLE_SIZE, start.y - HANDLE_SIZE, 
		2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
	    g.fillRect(end.x - HANDLE_SIZE, end.y - HANDLE_SIZE, 
		2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
	}
    }

    public boolean contains(int x, int y) {
	if (! constructed)
	    return false;
	if ((isSelected() && (startHandle.contains(x, y) ||
	    endHandle.contains(x, y))))
	    return true;

	double length1 = Math.sqrt((start.x - x) * (start.x - x) +
		(start.y - y) * (start.y - y));

	double length2 = Math.sqrt((end.x - x) * (end.x - x) +
		(end.y - y) * (end.y - y));
	return (Math.abs(length - length1 - length2) < 0.25);
    }

    public boolean contains(Point p) {
	return contains(p.x, p.y);
    }

    void setConstructed(boolean constructed) {
	this.constructed = constructed;
    }

    // Use for drawing handles only?
    public void setColorSource(ColorSource cs) {
    }

    public Object getModel() {
	return this;
    }

    public Component getView() {
	return this;
    }

    public void remove() {
        wb.remove(this);
        wb.componentRemoved(this);
    }

    public boolean isSelected() {
	return (state != STATE_UNSELECTED);
    }

    public void setSelected(boolean selected) {
	if (isSelected() != selected) {
	    if ((selected) && (wb.requestSelection(this))) {
		wb.setCurrentSelection(this);
		this.state = STATE_SELECTED;
		wb.requestEditor(this);
	    }
	    else {
		this.state = STATE_UNSELECTED;
		this.pointBeingMoved = null;
		if (wb.getCurrentSelection() == this) 
		    wb.setCurrentSelection(null);
		lastPoint = null;
	    }
	    repaint();
     	}
    }

    public void addPropertyChangeListener(PropertyChangeListener l) {
	listeners.addPropertyChangeListener(l);
    }
    public void removePropertyChangeListener(PropertyChangeListener l) {
	listeners.removePropertyChangeListener(l);
    }

    public synchronized Point getStart() {
	return start;
    }
    public synchronized void setStart(Point start) {
	Point orig = new Point(this.start);
	if (! orig.equals(start)) {
	    this.start = start;
 	    repaint();
	    listeners.firePropertyChange("start", orig, new Point(start));
	}
    }
    public synchronized Point getEnd() {
	return end;
    }
    public synchronized void setEnd(Point end) {
	Point orig = new Point(this.end);
	if (! orig.equals(end)) {
	    this.end = end;
 	    repaint();
	    listeners.firePropertyChange("end", orig, new Point(end));
	}
    }

    private boolean onHandle(Point endPoint, Point p) {
	return (Math.abs(endPoint.x - p.x) < 6) && 
		(Math.abs(endPoint.y - p.y) < 6);
    }

    public void mouseMoved(MouseEvent e) { }

// This needs to fire propChangeEvents. Note that WB.componentBoundsChanged
// is NOT appropriate, since the bounds of a line are always the same.
    public synchronized void mouseDragged(MouseEvent e) {
	if (isSelected()) {
	    if (lastPoint != null) {
		Point origStart = new Point(start);
		Point origEnd = new Point(end);
		int xDiff = e.getPoint().x - lastPoint.x;
		int yDiff = e.getPoint().y - lastPoint.y;
		if (state == STATE_RESHAPING) {
		    if (pointBeingMoved == start) {
		        start.x += xDiff;
		        start.y += yDiff;
	    	        listeners.firePropertyChange("start", origStart, 
			    new Point(start));
		    }
		    else if (pointBeingMoved == end) {
		        end.x += xDiff;
		        end.y += yDiff;
	    	        listeners.firePropertyChange("end", origEnd, 
			    new Point(end));
		    }
		}
		else if (state == STATE_MOVING) {
		    start.x += xDiff;
		    start.y += yDiff;
	    	    listeners.firePropertyChange("start", origStart, 
			new Point(start));
		    end.x += xDiff;
		    end.y += yDiff;
	    	    listeners.firePropertyChange("end", origEnd,
			new Point(end));
		}
	    }
	    lastPoint = e.getPoint();
	    repaint();
	}
    }

    public void mousePressed(MouseEvent e) {
   	setSelected(true);
	lastPoint = e.getPoint();
	if (isSelected()) {
	    if (onHandle(lastPoint, end) && !onHandle(lastPoint, start)) {
		state = STATE_RESHAPING;
		pointBeingMoved = end;
	    }
	    else if (onHandle(lastPoint, start) && !onHandle(lastPoint, end)) {
		state = STATE_RESHAPING;
		pointBeingMoved = start;
	    }
	    else
		state = STATE_MOVING;
	}
    }
    public void mouseReleased(MouseEvent e) {
	if (isSelected())
	    state = STATE_SELECTED;
	lastPoint = null;
	pointBeingMoved = null;
    }
    public void mouseClicked(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}

}


