package sieve.workbench;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.Serializable;

public class SimpleLinkComponent extends Component {

    private static final int ABASE 		= 10;
    private static final int ARROW_HEAD_SIZE 	= 25;

    Point 	destCenter	= null;
    Point 	sourceCenter	= null;
    int 	currX 		= 0;
    int 	currY 		= 0;
    Polygon 	polygon 	= null;
    Polygon 	oldPolygon 	= null;
    Color  	linkColor	= Color.black;

    Component source	= null;
    Component dest	= null;

    public SimpleLinkComponent(Component source) {
	this.source = source;
	setTempEnd(source.getLocation().x, source.getLocation().y);
    }

    public SimpleLinkComponent(Component source, Component dest) {
	this.source = source;
	this.dest = dest;
    }

    public void setColor(Color c) {
	linkColor = c;
	repaint();
    }

    public Color getColor() {
	return linkColor;
    }

    /**
     * Return the component from which the link was started.
     */
    public Component getSourceComponent() {
	return source;
    }
 
    /**
     * Return the component to which the link was drawn.
     */
    public Component getDestComponent() {
	return dest;
    }
    
    public void remove() {
	Container parent = getParent();
	if (parent != null)
	    parent.remove(this);
    }

    /**
     * Sets a "temporary" endpoint. This will be used for the endpoint
     * of the link until the destination component has been set.
     */
    public void setTempEnd(int x, int y) {
      	currX = x;
      	currY = y;
	repaint();
    }

    /**
     * Sets the destination component.
     */
    public void setDestComponent(Component dest) {
        this.dest = dest;
        repaint();
    }

    public void update(Graphics g) {
        paint(g);
    }

    private static int getDistance(Point p1, Point p2) {
	return (int) Math.sqrt(
	    (p1.x - p2.x) * (p1.x - p2.x) +
	    (p1.y - p2.y) * (p1.y - p2.y));
    }

    private static Point getIntersection(Rectangle bounds, Point lineStart, 
	Point lineEnd) {
	Point inside = null;
	Point outside = null;
	Point intersection = null;

	if (bounds.contains(lineEnd)) {
	    inside = lineEnd;
	    outside = lineStart;
	}
	if (bounds.contains(lineStart)) {
	    if (inside != null) { // Line is completely within the rect
		return null;
	    }
	    outside = lineEnd;
	    inside = lineStart;
	}
	if (inside == null) { // Line is completely outside the rect
	    return null;
	}

	if (inside.x == outside.x) { // Vertical line
	    if (outside.y <= bounds.y) // Outside pt is north
	         intersection = new Point(outside.x, bounds.y);
	    else
	         intersection = new Point(outside.x, bounds.y + bounds.height);
	}
	else if (inside.y == outside.y) { // Horizontal line
	    if (outside.x <= bounds.x) // Outside pt is west 
	         intersection = new Point(bounds.x, outside.y);
	    else
	         intersection = new Point(bounds.x + bounds.width, outside.y);
	}
	else {
	    double m = - (((double)inside.y - (double)outside.y)/
		((double)inside.x - (double)outside.x));
            double b = (double)inside.y - m * (double)inside.x;
	    double y =  m * (double)bounds.x + b;
	    if ((y >= bounds.y) && (y <= bounds.y + bounds.height)) { // e-w
		if (outside.x <= bounds.x) // Outside pt is west
	             intersection = new Point(bounds.x, 2 * bounds.y +
			bounds.height - (int)y);
		else
	             intersection = new Point(bounds.x + bounds.width, (int)y);
	    }
	    else { // n-s line
		double x = ((double)bounds.y - b)/m;
	        if (outside.y <= bounds.y) // Outside pt is north
	            intersection = new Point(2 * bounds.x + bounds.width -
			(int)x, bounds.y);
		else
	            intersection = new Point((int)x, bounds.y + bounds.height);
	    }
	}

	return intersection;
    }

    public void paint(Graphics g) {
    	if (source == null){
      	    erase(g);
	    getParent().remove(this);
	    return;
        }
        Dimension mySize = getSize();

	Point sourceLoc = source.getLocation();
	Dimension sourceSize = source.getSize();
	Rectangle srcBounds = source.getBounds();

	srcBounds.x += srcBounds.width/2; // now it has the center x
	srcBounds.y += srcBounds.height/2; // now it has the center y

	sourceCenter = new Point(sourceLoc.x + (sourceSize.width/2), 
			     sourceLoc.y + (sourceSize.height/2));
			     
	Point destLoc = new Point(currX, currY);
	Dimension destSize = new Dimension(2, 2);

	Rectangle destBounds = null;
	if (dest != null) {
            destLoc = dest.getLocation();
            destSize = dest.getSize();
	    destBounds = dest.getBounds();
        }
        destCenter = new Point(destLoc.x + (destSize.width/2), 
			   destLoc.y + (destSize.height/2));
	if (destBounds == null)
            destBounds = new Rectangle(destCenter.x, destCenter.y, 2, 2);

	Point srcIntersection = getIntersection(source.getBounds(),
	    destCenter, sourceCenter);
	if (srcIntersection == null) 
	    return;

	Point destIntersection = getIntersection(destBounds,
	    destCenter, srcIntersection);
	if (destIntersection == null) 
	    return;

	Point arrowStart = null;
	int   dist = getDistance(destIntersection, srcIntersection);
	if (dist < 2 * ARROW_HEAD_SIZE)
	    arrowStart = new Point((destIntersection.x + srcIntersection.x)/2,
	    	(destIntersection.y + srcIntersection.y)/2);
	else 
	    arrowStart = new Point(destIntersection.x - 
		ARROW_HEAD_SIZE * (destIntersection.x-srcIntersection.x)/dist,
	    	destIntersection.y - 
		ARROW_HEAD_SIZE * (destIntersection.y-srcIntersection.y)/dist);

        double theta;
        if (arrowStart.y == destIntersection.y)
            theta = 0;
        else
            theta = Math.atan((destIntersection.x - arrowStart.x)/
                (double)(arrowStart.y - destIntersection.y)) + Math.PI/2;

        // Copied from Andy
        int[] xArray = new int[5];
        xArray[0] = arrowStart.x;
        xArray[1] = arrowStart.x - (int)(ABASE * Math.sin(theta));
        xArray[2] = destIntersection.x;
        xArray[3] = arrowStart.x + (int)(ABASE * Math.sin(theta));
        xArray[4] = arrowStart.x;
        int[] yArray = new int[5];
        yArray[0] = arrowStart.y;
        yArray[1] = arrowStart.y + (int)(ABASE * Math.cos(theta));
        yArray[2] = destIntersection.y;
        yArray[3] = arrowStart.y - (int)(ABASE * Math.cos(theta));
        yArray[4] = arrowStart.y;
        polygon = new Polygon(xArray, yArray, xArray.length);
 
   
        // Erase the old polygon
        // Can't just do a clearRect 'cause it would wipe out other links, too
        if (oldPolygon != null) {
            erase(g);
        }

        // Draw the polygon
        g.setColor(linkColor);
	g.drawLine(srcIntersection.x, srcIntersection.y, arrowStart.x, 
	    arrowStart.y);
        g.fillPolygon(polygon);
    
        // Set the oldPolygon for later erasure
        oldPolygon = polygon;
    }

    public void erase(Graphics g) {
        g.setColor(getParent().getBackground());
        g.fillPolygon(oldPolygon);
    }

    private static void drawWideLine(Graphics g, int width, int x1, 
	int y1, int x2, int y2) {
    	for(int counter = 0; counter < width; counter++) {
      	    g.drawLine(x1+counter, y1+counter, x2, y2);
    	}
    }

    public boolean contains(int x, int y) {
    	if (polygon != null)
      	    return polygon.contains(x, y);
    	else
      	    return false;
    }

    public Dimension minimumSize() {
        return getSize();
    }
    public Dimension preferredSize() {
        return minimumSize();
    }

}


