package sieve.workbench;

import java.util.Hashtable;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;

import com.sun.java.swing.*;

import sieve.workbench.event.*;

public class Workbench extends JLayeredPane {
    private Tool 			currentTool = null;
    private boolean 			toolIsRetained = false;
    private WorkbenchComponent 		currentSelection = null;
    private Link			currentLink = null;
    private WorkbenchComponentEditor 	editor = null;
    private Image 			offscreen = null;

    private ToolEventSupport		toolListeners = null;
    private SelectionListenerSupport	selectionListeners = null;
    private WorkbenchListenerSupport	workBenchListeners = null;

    // While we're at it, we may as well have the workbench act like
    // a bean as well. This is a little redundant with ToolListeners,
    // but that's probably OK.
    private PropertyChangeSupport       propListeners = null;

    // We'll track the last edits (EditEvents, for now) for
    // each component. Simultaneous duplicate changes will be ignored.
    private Hashtable			lastEdits = null;


    public Workbench() {
	setLayout(null);
	paintBackground = true; // Protected field in JLayeredPane
	setBackground(Color.white);
	addMouseListener(new WorkbenchMouseListener(this));
	toolListeners = new ToolEventSupport(this);
	selectionListeners = new SelectionListenerSupport(this);
	workBenchListeners = new WorkbenchListenerSupport(this);
 	propListeners = new PropertyChangeSupport(this);
	lastEdits = new Hashtable();
    }

    /**
     * Sets the current editor for WorkbenchComponents. When a component
     * requests editing, the request will be forwarded to this editor.
     */
    public void setWorkbenchComponentEditor(WorkbenchComponentEditor e) {
	if (e != editor) {
	    WorkbenchComponentEditor old = this.editor;
	    this.editor = e;
	    propListeners.firePropertyChange("workbenchComponentEditor", 
		old, e);
	}
    }
    /**
     * Returns the current editor for WorkbenchComponents. When a component
     * requests editing, the request will be forwarded to this editor.
     */
    public WorkbenchComponentEditor getWorkbenchComponentEditor() {
	return editor;
    }
    

    /**
     * Sets the current Tool. All MouseEvents will be forwarded to 
     * this tool until it returns a Wrapper around a new component
     * or throws an exception.
     */
    public synchronized void setCurrentTool(Tool t) {
	Tool oldTool = getCurrentTool();
	if (currentTool != null) {
	     removeMouseListener(currentTool);
	     removeMouseMotionListener(currentTool);
	}
	currentTool = t;
	if (currentTool == null) {
	    setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	    requestFocus();
	}
	else {
	    addMouseListener(currentTool);
	    addMouseMotionListener(currentTool);
	    Cursor c = currentTool.getPreferredCursor();
	    if (c != null)
		setCursor(c);
	    else
		setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
	}
	toolListeners.toolChanged(oldTool);
	propListeners.firePropertyChange("currentTool", oldTool, currentTool);
    }

    /**
     * Returns the current Tool. If this is non-null, then all MouseEvents
     * will be forwarded to the Tool until it returns an object.
     */
    public Tool getCurrentTool() {
	return currentTool;
    }

    /**
     * Sets whether the current tool should be retained after a component is
     * added.
     */
    public void setToolRetained(boolean toolIsRetained) {
	if (this.toolIsRetained != toolIsRetained) {
	    this.toolIsRetained = toolIsRetained;
	    toolListeners.toolRetainedStateChanged();
	    propListeners.firePropertyChange("toolRetained", 
		new Boolean(!toolIsRetained), new Boolean(toolIsRetained));
	}
    }

    /**
     * Returns true if the current tool will be retained after a component is
     * added.
     */
    public boolean isToolRetained() {
	return toolIsRetained;
    }

    /**
     * Registers a listener for tool events (tool set, tool used, tool unset)
     */ 
    public void addToolListener(ToolListener tl) {
	toolListeners.addToolListener(tl);
    } 
 
    /**
     * Removes a listener for tool events (tool set, tool used, tool unset)
     */
    public void removeToolListener(ToolListener tl) {
	toolListeners.removeToolListener(tl);
    }

    /**
     * Called by Tool objects when they have successfully created an object.
     * Note that the Workbench does not add the object to itself. This method
     * simply allows the Workbench to notify listeners that they new object
     * exists. 
     * Workbench listeners are always notified, but tool listeners are
     * only notified if the tool used to create the component is the
     * "current" tool.
     * If the "sticky" state of the tool is false, the current tool is
     * set to null <I>after</I> all listener notifications have finished.
     *
     * @param wbc The component that was added
     * @param tool The tool which created the component
     */
    public void componentAdded(WorkbenchComponent wbc, Tool tool) {
	// Workbench listeners find out regardless
	workBenchListeners.componentAdded(wbc);

	// Only notify tool listeners if this was created by the user
	// interacting with the "current" tool.
	if (tool == currentTool) {
	    toolListeners.toolUsed(wbc);
	    if (! toolIsRetained)
	         setCurrentTool(null);
	    // Need to be careful with this -- we don't want components
	    // created by remote users to be selected upon creation.
	    // setCurrentSelection(wbc);
	}
    }

    /**
     * Called when a WorkbenchComponent has removed itself.
     * Note that the Workbench does not add the object to itself. 
     * This method simply allows the Workbench to notify listeners that 
     * the object has been removed.
     * <P><I><B>The removal process needs alot more thought</B></I>
     */
    public void componentRemoved(WorkbenchComponent wbc) {
	if (wbc == getCurrentSelection())
	    setCurrentSelection(null);
	repaint();
	workBenchListeners.componentRemoved(wbc);
    }

    /**
     * Adds a WorkbenchListener.
     * Workbench listeners receive notification when WorkbenchComponents
     * are added, removed, moved, resized, linked, unlinked, or edited.
     */
    public void addWorkbenchListener(WorkbenchListener l) {
	workBenchListeners.addWorkbenchListener(l);
    }
    /**
     * Removes a WorkbenchListener.
     */
    public void removeWorkbenchListener(WorkbenchListener l) {
	workBenchListeners.removeWorkbenchListener(l);
    }

    /**
     * Adds a SelectionListener. Selection listeners receive notification 
     * when WorkbenchComponents request selection and may veto such requests.
     * <P>
     * This functionality is provided to support floor control mechanisms.
     */
    public void addSelectionListener(SelectionListener l) {
	selectionListeners.addSelectionListener(l);
    }

    /**
     * Removes a SelectionListener.
     */
    public void removeSelectionListener(SelectionListener l) {
	selectionListeners.removeSelectionListener(l);
    }

    /**
     * Adds a PropertyChangeListener. These will be notified when
     * the following properties change: currentTool, toolIsRetained,
     * currentSelection, currentComponentEditor.
     */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
	propListeners.addPropertyChangeListener(listener);
    }

    /**
     * Removes a PropertyChangeListener.
     */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
	propListeners.removePropertyChangeListener(listener);
    }

    /**
     * Sets the current selection to the specified WorkbenchComponent.
     * If another component is already selected, it is first 
     * deselected.
     */
    public void setCurrentSelection(WorkbenchComponent c) {
	if (c != currentSelection) {
	    WorkbenchComponent old = currentSelection;
	    if (currentSelection != null) {
		currentSelection.setSelected(false);
		workBenchListeners.componentDeselected(currentSelection);
	    }
	    currentSelection = c;
	    if (currentSelection != null) {
		currentSelection.setSelected(true);
		workBenchListeners.componentSelected(currentSelection);
	    }
	    else if (editor != null)
		editor.setTarget(null);
	    propListeners.firePropertyChange("currentSelection",
		old, currentSelection);
	}
    }

    /**
     * Returns a reference to the current selection. This will be null
     * if no component is currently selected.
     */
    public WorkbenchComponent getCurrentSelection() {
	return currentSelection;
    }

    /**
     * Called by a WorkbenchComponent to request selection. All of the
     * Workbench's SelectionListeners are first given an opportunity to
     * veto the request (e.g., a floor control mechanism could decide that
     * the current user does not have privileges to change the requested
     * objects.)
     * <P>
     * If there is currently an active, unfinished link, the link 
     * is given an opportunity to use the requesting component as  
     * its destination. If there is an active link, then the
     * component's request is denied regardless of whether the link
     * accepts the component or not.
     * <P>
     * Well-behaved
     * WorkbenchComponents must <B>always</B> call this prior to setting
     * themselves as the current selection.
     * <P>
     * <I>Should this also
     * send the MouseEvent that initiated the request for multiple-selection
     * support? Should we wrap it in a more intuitive way for link 
     *  requests?</I>
     */
    public boolean requestSelection(WorkbenchComponent w) {
	if (!selectionListeners.selectionApproved(w)) 
	    return false;
	if (currentLink != null) {
	    try {
	        currentLink.setDestComponent(w);
	        removeMouseMotionListener(currentLink);
		workBenchListeners.linkAdded(currentLink);
		currentLink = null;
	    }
	    catch(LinkException le) {
		System.out.println("Link Rejected: " + le.getMessage());
	    }
	    return false;
  	}
	return true;
    }

    /**
     * This method should be called by any outside entity which adds
     * link to the Workbench. (e.g., a WorkbenchComponent or a 
     * collaboration support object such as sieve.workbench.collab.Controller) 
     * Links themselves should NOT call this method, or bad things can
     * happen when we're in collaborative mode. <I>(This restriction can
     * probably be removed by checking more carefully in Broadcaster.)</I>
     */
    public void linkAdded(Link link) {
	workBenchListeners.linkAdded(link);
    }

    /**
     * Called to request that the given workbench component be edited using
     * the current editor.
     */
    public boolean requestEditor(WorkbenchComponent w) {
	if ((getCurrentSelection() == w) && (editor != null)) {
	    editor.setTarget(w);
	    return true;
	}
	return false;
    }

    /**
     * Called by a wrapper to request that a link be started. The request
     * will be denied if, for example, there is already a link attempt
     * in progress.
     * The WorkbenchComponent does not have to be the current selection,
     * but if it isn't the currently selected component is deselected.
     * The current tool is also deselected, to prevent a tool from 
     * trying to create a new component based on whatever the user does
     * to make a link.
     */
    public boolean requestLinkStart(WorkbenchComponent w, Link l) {
	if (currentLink == null) {
	    if (getCurrentSelection() != w)
		setCurrentSelection(null);
	    currentLink = l;
	    setCurrentTool(null);
	    addMouseMotionListener(currentLink);
	    workBenchListeners.linkStarted(l);
	    return true;
	}
	return false;
    }

    /**
     * Called by a link to notify the workbench that it has been removed.
     */
    public void linkRemoved(Link l) {
	if (currentLink == l) {
	    removeMouseMotionListener(currentLink);
	    currentLink = null;
	}
	workBenchListeners.linkRemoved(l);
    }

    /**
     * Called by a WorkbenchComponent after it has finished moving or 
     * resizing itself.
     * <P>
     * <I>This is not just a change to the "bounds" property of the 
     * component, since the bounds may not be a property of the object
     * represented by the WorkbenchComponent, but rather of its visual
     * component.</I>
     */
    public void componentBoundsChanged(WorkbenchComponent w) {
	workBenchListeners.componentBoundsChanged(w);
	repaint(); // Make sure that any LW components (eg, links) are drawn
    }

    /**
     * Called by a WorkbenchComponent after the thing returned by 
     * getModel() has been edited. PropertyChangeEvent should (eventually)
     * be extended to include additional information if the 
     * WorkbenchComponent is an aggregate.
     */
    public synchronized void componentEdited(WorkbenchComponent w, 
	EditEvent e) {
	// We attempt to avoid sending duplicate notifications. This
	// can happen easily, since a change to a bound property made via
	// the (external) property editor will be detected by the 
	// WorkbenchComponent, and will be indistinguishable from a change
	// made via the components customizer.
	EditEvent last = (EditEvent) lastEdits.get(w);
	if ((last == null) || (! last.isEquivalentTo(e))) {
	    lastEdits.put(w, e);
	    workBenchListeners.componentEdited(w, e);
	}
    }

    /**
     * Convenience method. Most edits will be property changes. Sieve-savvy
     * components which produce more complex EditEvents should call the other
     * version of this method.
     */
    public void componentEdited(WorkbenchComponent w, PropertyChangeEvent e) {
	componentEdited(w, new PropertyEditEvent(e));

	// String hash = e.getPropertyName() + e.getNewValue();
	// synchronized (lastEdits) {
	//     String last = (String) lastEdits.get(w);
	//     if ((last != null) && last.equals(hash)) 
	//	    return;
	//    lastEdits.put(w, hash);
	// }
	// workBenchListeners.componentEdited(w, e);
    }

    /**
     * Used by the Workbench's mouse listener class to remove the current
     * link if the user begins a link and then clicks on the workspace.
     */
    private void clearCurrentLink() {
	if (currentLink != null) {
	    currentLink.remove();
	    currentLink = null;
	    repaint();
	}
    }

    public void update(Graphics g) {
	paint(g);
    }

    public void paint(Graphics g) {
	Dimension size = getSize();
	if ((size.width != 0) && (size.height != 0)) {
	    if ((offscreen == null) || (offscreen.getHeight(null)!=size.height)
	        || (offscreen.getWidth(null) != size.width)) {
	        offscreen = createImage(size.width, size.height);
	    }
	    Graphics og = offscreen.getGraphics();
	    og.setClip(0, 0, size.width, size.height);
	    og.clearRect(0, 0, size.width, size.height);
	    super.paint(og);
	    g.drawImage(offscreen, 0, 0, null);
	    og.dispose();
	}
    }

    public void print(Graphics g) {
	g.setClip(0, 0, getSize().width, getSize().height);
	super.print(g);
	paint(g);
    }

    /**
     * Wrapper around <CODE>Component.processEvent</CODE> used to allow
     * collaboration support objects to enqueue AWT events.
     */
    public void injectEvent(AWTEvent e) {
	processEvent(e);
    }

    private class WorkbenchMouseListener extends MouseAdapter {
	Workbench w = null;
	public WorkbenchMouseListener(Workbench w) {
	    this.w = w;
 	}
	public void mouseClicked(MouseEvent e) {
	    if (w.getCurrentTool() == null) {
		w.setCurrentSelection(null);
	    }
	    w.clearCurrentLink();
	    w.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	    w.requestFocus();
 	}
    }

}



