package sieve.workbench.collab;

import sieve.workbench.event.*;
import sieve.workbench.collab.msg.*;
import sieve.workbench.*;
import java.beans.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/**
 * The Broadcaster listens for interesting things on the Workbench. When
 * something interesting happens, an appropriate message (from the msg
 * subpackage) is created and handed to the Workbench client to be sent.
 */
public class Broadcaster implements WorkbenchListener, 
    SelectionListener, ToolListener {

    private MouseEventRecorder recorder = null;
    private Workbench wb = null;
    private ComponentRegistry compRegistry = null; 
    private Controller myController = null;

    private WorkbenchClient client = null;

    /**
     * A controller is a "virtual" user of a workbench. If our client is
     * null, we will deliver interesting events to this controller.
     * this can be used to create a "mirror" workbench for testing.
     */
    private Controller controller = null;

    /**
     * Construct a new Broadcaster. For testing, passing a null client but
     * a non-null controller will cause messages to be delivered directly
     * to the controller. This allows creation of a local "mirror" workbench.
     *
     * @param wb The Workbench to listen to
     *
     * @param controller The controller which is driving the workbench. This is consulted before sending some messages to prevent looping (rebroadcasting messages based on changes by remote users.)
     * 
     * @param client A client object which knows how to send messages.
     */
    public Broadcaster(Workbench wb, Controller controller, 
	WorkbenchClient client) {
	this.wb = wb;
	this.controller = controller;
	this.client = client;

	recorder = new MouseEventRecorder();
	wb.addMouseListener(recorder);
	wb.addMouseMotionListener(recorder);

	compRegistry = ComponentRegistry.getComponentRegistry(wb);
    }

    /**
     * The Broadcaster should not begin listening to the workbench
     * until all "late joiner" messages have been received. It looks
     * like the Controller will have to determine this and call this
     * method.
     */
    public void startBroadcasting() {
	wb.addWorkbenchListener(this);
        wb.addSelectionListener(this);
        wb.addToolListener(this);
    }

    public void stopBroadcasting() {
	wb.removeWorkbenchListener(this);
        wb.removeSelectionListener(this);
        wb.removeToolListener(this);
    }

    // TEMP TEMP TEMP
    private void print(String msg) {
	//System.out.println("--> CollabContoller UNIMPLEMENTED: " + msg);
    }

    // ------------- WorkbenchListener methods -------------
    /**
     * Not actually used. Component adding is handled in the toolUsed(...)
     * method.
     */
    public void componentAdded(WorkbenchComponentEvent e) {
    }

    /**
     * This will eventually be used (?) to implement component locking.
     */
    public void componentSelected(WorkbenchComponentEvent e) {
            print("componentSelected " +
                e.getWorkbenchComponent().getModel().toString());
    }
    /**
     * This will eventually be used (?) to implement component (un)locking.
     */
    public void componentDeselected(WorkbenchComponentEvent e) {
            print("componentDeselected " +
                e.getWorkbenchComponent());
    }

    public void componentRemoved(WorkbenchComponentEvent e) {
	Message m = new RemoveComponentMessage(
	    e.getWorkbenchComponent().getUniqueID());
	if (client == null)
	    m.deliver(controller);
	else 
	    client.send(m);
    }
    public void componentBoundsChanged(WorkbenchComponentEvent e) {
	Rectangle b = controller.getCurrentBounds();
	Rectangle newBounds = 
	    e.getWorkbenchComponent().getView().getBounds();

	Message m = new BoundsMessage(e.getWorkbenchComponent().getUniqueID(),
	    newBounds);

	if (client == null)
	    m.deliver(controller);
	else if ((b == null) || !(b.equals(newBounds)))
	    client.send(m);
    }
    public void componentZOrderChanged(WorkbenchComponentEvent e) {
            print("componentZOrderChanged " +
                e.getWorkbenchComponent().getModel().toString());
    }
    public void componentEdited(WorkbenchComponentEvent e) {
	synchronized(controller) {
	    EditEvent cur = controller.getCurrentEditEvent();
	    if ((cur == null) || (! cur.isEquivalentTo(e.getEditEvent()))) {
	        Message m = null;
	        // Need something like a transient interface that WBC's can imp
	        if (e.getWorkbenchComponent().getModel() instanceof Participant)
		    m = new TransientPropertyChangeMessage(
	                e.getWorkbenchComponent().getUniqueID(), 
	                e.getPropertyChangeEvent());
	        else
		    m = new EditMessage(
	                e.getWorkbenchComponent().getUniqueID(), 
	                e.getEditEvent());
		client.send(m);
	    }
	}
    }
 
    /**
     * Not currently used by Broadcaster.
     */
    public void linkStarted(LinkEvent e) {
    }

    public void linkAdded(LinkEvent e) {
	Link link = e.getLink();

	// If its not in our registry, then it must have been added by
	// the user (and not the Controller).
	if (compRegistry.getLink(link.getUniqueID()) == null) {
	    Message m = new MakeLinkMessage(
		link.getClass(),
		link.getSourceComponent().getUniqueID(), 
		link.getDestComponent().getUniqueID(), 
		link.getUniqueID(),
		link.getState());
	    if (client == null)
	        m.deliver(controller);
	    else
	        client.send(m);
	    compRegistry.registerLink(link);
	}
    }
    public void linkRemoved(LinkEvent e) {
	Link link = e.getLink();

	// If it's still in our registry, then the user (and not the
	// Controller) must have removed it.
	if (compRegistry.getLink(link.getUniqueID()) != null) {
	    Message m = new RemoveLinkMessage(link.getUniqueID());
	    if (client == null)
	        m.deliver(controller);
	    else
	        client.send(m);
	    compRegistry.removeLink(link);
	}
    }
    // -----------------------------------------------------

    // ------------- SelectionListener methods -------------
    /** 
     * For now, only allow selection of components which appear in
     * the registry. This will prevent the user from inadvertantly
     * selecting components which are being constructed.
     * Eventually if we implement locking, we will want to check here
     *  to see if another user has locked the component.
     */
    public boolean selectionApproved(WorkbenchComponent c) {
	// PLI -- this breaks circuit sim if local recording is present :-(
        // return (compRegistry.getWorkbenchComponent(c.getUniqueID()) != null);
	return true;
    }
    // -----------------------------------------------------

    // --------------- ToolListener methods ----------------
    public void toolChanged(ToolEvent e) {
            print("toolChanged: " + e.toString());
	if (e.getCurrentTool() == null)
	    recorder.stopRecording();
	else
	    recorder.startRecording();
    }

    /**
     * If we get toolUsed, we know that a component has been created by
     * the local user, and not by a remote user. (componentAdded will be
     * called in either case.)
     */
    public void toolUsed(ToolEvent e) {
        print("toolUsed: " + e.toString());
	compRegistry.registerWorkbenchComponent(e.getCreatedComponent());
	MouseEvent[] evts = recorder.stopRecording();
	Message m = new AddComponentMessage(
	    e.getCreatedComponent().getUniqueID(),
	    e.getCreatedComponent().getModel().getClass(),
	    evts);
	if (client == null)
	    m.deliver(controller);
	else
	    client.send(m);
	if (wb.isToolRetained())
	    recorder.startRecording();
    }
    public void toolRetainedStateChanged(ToolEvent e) {
            print("toolRetainedStateChanged: " + e.toString());
    }
    // -----------------------------------------------------


    // -----------------------------------------------------
    /**
     * MouseEventRecorder simply records Mouse and MouseMotion events.
     * The recorder is turned on when the user begins to create a component,
     * and off when the user finishes. The sequence of mouse event is then
     * sent to remote users, where it is "re-played" to re-create the
     * component.
     */
    private class MouseEventRecorder implements MouseListener, 
	MouseMotionListener{
	private Vector events = new Vector();
	private boolean recording = false;
	private boolean firstPressRecorded = false;

	private void recordEvent(MouseEvent e) {
	    if ((recording) && (firstPressRecorded)) {
	  	synchronized(this) {
	     	    events.addElement(e);
		}
	    }
 	}

        public synchronized void startRecording() {
	    events.removeAllElements();
	    recording = true;
	    firstPressRecorded = false;
	}

        public boolean isRecording() {
	    return recording;
	}

        public synchronized MouseEvent[] stopRecording() {
	    recording = false;
	    firstPressRecorded = false;
	    MouseEvent[] evs = new MouseEvent[events.size()];
	    events.copyInto(evs);
	    events.removeAllElements();
	    return evs;
	}

	public void mouseClicked(MouseEvent e) {
	    recordEvent(e);
	}
	public void mousePressed(MouseEvent e) {
	    firstPressRecorded = true;
	    recordEvent(e);
	}
	public void mouseReleased(MouseEvent e) {
	    recordEvent(e);
	}
	public void mouseEntered(MouseEvent e) {
	    recordEvent(e);
	}
	public void mouseExited(MouseEvent e) {
	    recordEvent(e);
	}
	public void mouseMoved(MouseEvent e) {
	    recordEvent(e);
	}
	public void mouseDragged(MouseEvent e) {
	    recordEvent(e);
	}

    }

}
