package sieve.workbench.collab;

import sieve.workbench.*;
import sieve.workbench.collab.msg.Message;
import vt.cs.widgets.*;
import java.awt.Color;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import java.awt.event.*;

/**
 * This class does the dirty work of attaching a workbench to a collaborative
 * session. To use this class:
 * <OL>
 * <LI> Instantiate an instance of this class
 * <LI> Call getRadarPane() if you want to display the workbench in a 
 *      radarpane
 * <LI> Call restoreWorkbench() if you want to bring the workbench up-to-date
 * </OL>
 * <P>
 * If you want to do manually what this class does for you, you need:
 * <UL>
 * <LI> The host on which the server is running
 * <LI> The name of the session 
 * <LI> The name of the channel within that session 
 * </UL>
 * Then:
 * <UL>
 * <LI> Create the workbench
 * <LI> Get a WorkbenchChannelManager using host and channel name
 * <LI> Tell the WorkbenchChannelManager to add the channel (no effect if
 *      the channel already exists
 * <LI> Instantiate a new Controller to drive the workbench
 * <LI> Instantiate a new JSDAWorkbenchClient to handle client-side 
 *      JSDA interaction
 * <LI> Instantiate a new Broadcaster to listen to the workbench.
 * <LI> If using RadarPane:
 *      <UL>
 *      <LI> Tell the broadcaster to start broadcasting
 *      <LI> Use ParticipantTool.getLocalParticipant(...) to get a 
 *           WorkbenchParticipant
 *      <LI> Create an MURadarPane giving it the WorkbenchParticipant
 *      <LI> Tell the ParticipantTool about the radarpane using 
 *           ParticipantTool.setMURadarPane(...)
 *      <LI> Add the workbench to the radar pane 
 *      <LI> Do any other radar pane configuration
 *      </UL>
 * <LI> If restoring workbench state:
 *      <UL>
 *      <LI> Make sure that the workbench is added to a component and displayed
 *      <LI> Tell the Broadcaster to stopBroadcasting()
 *      <LI> The JSDAWorkbenchClient should not be delivering messages to the
 *        controller. You can check this by seeing if isOKToDeliverMessages()
 *        returns false. setOKToDeliverMessages(false) if you want to be safe.
 *      <LI> Use the WorkbenchChannelManager to getStoredMessages(...)
 *      <LI> Deliver each message to the controller
 *      </UL>
 * <LI> Tell the Broadcaster to startBroadcasting()
 * <LI> Tell the JSDAWorkbenchClient to start delivering messages to the 
 *      controller: client.setOKToDeliverMessages(true)
 * </UL>
 */
public class CollaborationSupport {
    private Workbench wb = null;
    private String host = null;
    private String session = null;
    private String channel = null;
    private String user = null;
    private Color userColor = Color.black;

    private WorkbenchChannelManager mgr = null;
    private Controller controller = null;
    private JSDAClient client = null;
    private Broadcaster broadcaster = null;

    private WorkbenchParticipant participant = null;
    private MURadarPane pane = null;

    /**
     * Returns the names of all available sessions on the specified host.
     */
    public static String[] getSessions(String host) {
	String[] names = new String[0];
        try {
	    Registry r =
                LocateRegistry.getRegistry(host, Server.REGISTRY_PORT);
	    ServerInterface server = (ServerInterface) r.lookup(Server.NAME);
            names = server.getSessions();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        return names;
    }

    /**
     * Returns the names of all available channels on the specified session
     * on the specified host.
     */
    public static String[] getChannels(String host, String session) {
	String[] names = new String[0];
        try {
	    Registry r =
                LocateRegistry.getRegistry(host, Server.REGISTRY_PORT);
	    ServerInterface server = (ServerInterface) r.lookup(Server.NAME);
            names = server.getChannels(session);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        return names;
    }

    /**

    /**
     * Constructs a new CollaborationSupport object. The host parameter
     * may be null if you still want radarpane support, but aren't
     * connecting to a server.
     */
    public CollaborationSupport(Workbench wb, String host, String session, 
	String channel, String user, Color userColor) {
	this.wb = wb;
	this.host = host;
	this.session = session;
	this.channel = channel;
	this.user = user;
	this.userColor = userColor;

	if (host != null) {
	    // Get a manager to handle channel creation and interaction
	    mgr = WorkbenchChannelManager.getManager(host, session);

	    // Make sure the specified channel exists
	    mgr.addChannel(channel);

	    // Make a Controller to drive the workbench from incoming msgs
	    controller = new Controller(wb);

	    // Make a client to handle JSDA input and output
	    client = 
		new JSDAClient(mgr.getSession(), channel, controller);

	    // Make a broadcaster to listen to the workbench and generate
	    // messages representing interesting things that happen there
	    broadcaster = new Broadcaster(wb, controller, client);

	    // Make sure it's broadcasting when we create our local participant
	    // in the next step
            broadcaster.startBroadcasting();
	}

	// Make a representation of the current user
	participant = ParticipantTool.getLocalParticipant(wb);
	participant.setName(user);
	participant.setColor(userColor);


	// If the workbench is already showing, then we can restore it.
	// Otherwise we'd like to just be notified when it is shown, but
	// this doesn't seem to work (in 1.1.2, anyway). It is therefore
	// necessary to call restoreWorkbench() manually in some cases.
	if (wb.isShowing())
	    restoreWorkbench();
	else 
	    wb.addComponentListener(new CSWBCompListener(wb, this));
    }

    /**
     * Restores the state of the workbench based on stored messages for
     * the channel, then tell the client it can start delivering messages
     * that came in over JSDA to the controller. This method has no effect if 
     * we weren't given a host (ie, aren't really collaborating).
     */
    public synchronized void restoreWorkbench() {
	if (host != null) {
	    // All hell breaks loose if we start driving the workbench
	    // with old messages without turning off broadcasting.
	    broadcaster.stopBroadcasting();

	    // Get all of the old messages
	    Message[] msgs = mgr.getStoredMessages(channel);
            for (int i = 0; i < msgs.length; i++) 
                msgs[i].deliver(controller);

	    // Restart broadcasting
            broadcaster.startBroadcasting();
            client.setOKToDeliverMessages(true);
	}
    }

    /**
     * Returns a RadarPane containing our Workbench. If the Workbench was
     * already contained in something it is removed and added to the
     * RadarPane.
     */ 
    public synchronized MURadarPane getRadarPane() {
	if (pane == null) {
	    if (wb.getParent() != null)
		wb.getParent().remove(wb);
	    pane = new MURadarPane(RadarPane.SCROLLBARS_AS_NEEDED, participant);
	    wb.addWorkbenchListener(new RadarPaneUpdater(pane));
	    ParticipantTool.setMURadarPane(wb, pane);
	    pane.add(wb);
	    pane.setDrawCursors(true);
	    pane.setDrawLabels(true);
	}
	return pane;
    }

    /**
     * Logs out of the server and cleans up all network connections.
     */
    public void shutdown() {
	ParticipantTool.shutdown(wb);
	if (client != null)
	    client.shutdown();
    }

    // This appears to be non-funcational :-(
    private class CSWBCompListener extends ComponentAdapter {
	private Workbench wb = null;
	private CollaborationSupport cs = null;
	public CSWBCompListener(Workbench wb, CollaborationSupport cs) {
	    this.wb = wb;
	    this.cs = cs;
	}

	public void componentShown(ComponentEvent e) {
	    wb.removeComponentListener(this);
	    cs.restoreWorkbench();
	}
    }

}



