package sieve.workbench.collab;

import java.util.*;
import java.lang.reflect.*;
import java.beans.*;
import java.awt.*;
import java.awt.event.*;
import sieve.workbench.*;
import sieve.workbench.event.*;
import java.io.Serializable;

/**
 * Implementation of the WorkbenchController interface which can 
 * drive a Workbench object. 
 *
 * @see JSDAWorkbenchClient
 */
public class Controller implements WorkbenchController {
    private Workbench wb = null;
    private ComponentRegistry compRegistry = null;

    private Rectangle currentBounds = null;
    private Lock currentBoundsLock = null;

    private EditEvent currentEdit = null;
    private Lock currentEditLock = null;

    private Link currentLink = null;
    private Lock currentLinkLock = null;

    public Controller(Workbench wb) {
	this.wb = wb;
	compRegistry = ComponentRegistry.getComponentRegistry(wb);
	currentLinkLock = new Lock();
	currentBoundsLock = new Lock();
	currentEditLock = new Lock();
    }

    public void doAddComponent(String className, MouseEvent[] evts, 
	UniqueID id) {
	if (compRegistry.getWorkbenchComponent(id) == null) {

            Class cl = null;
            // This is a problem. We try Class.forName first to see if the
            // system classloader can find this class. Otherwise our only
            // other option is to try to Jar loader. There are probably
            // cases where neither of these options will work. Hmmmmmm....
            try {
                cl = Class.forName(className);
            }
            catch(Exception e) {
                try {
                 cl = sieve.workbench.jar.SimpleClassLoader.ourLoader.loadClass(
                     className);
                }
                catch(Exception ex) {
                    ex.printStackTrace();
                }
            }

	    IDSetter ids = new IDSetter(wb, cl, id, this);
	    Tool tool = ToolFactory.getToolFactory().getTool(cl, wb);
	    MouseEventDelegator med = new MouseEventDelegator(tool, tool, wb);
	    for (int i = 0; i < evts.length; i++)
                med.fireEvent(evts[i]);
	}
    }

    /**
     * IDSetter objects listener for a component to be added to a
     * workbench. When a component is added with a matching class,
     * the id of the component is set to the specified unique id,
     * the component is registered in the component registry, and
     * and the Controller is notified.
     */
    private class IDSetter extends WorkbenchListenerAdapter {
	private Workbench wb = null;
	private Class cl = null;
	private UniqueID id = null;
	private Controller c = null;

	public IDSetter(Workbench wb, Class cl, UniqueID id, Controller c) {
	    this.wb = wb;
	    this.cl = cl;
	    this.id = id;
	    this.c = c;
	    wb.addWorkbenchListener(this);
	}
	public void componentAdded(WorkbenchComponentEvent e) {
	    WorkbenchComponent wbc = e.getWorkbenchComponent();
	    if (wbc.getModel().getClass().equals(cl)) {
		synchronized (this) {
		    wbc.setUniqueID(id);
		    ComponentRegistry r = 
			ComponentRegistry.getComponentRegistry(wb);
		    r.registerWorkbenchComponent(wbc);
	            wb.removeWorkbenchListener(this);
		}
		synchronized (c) {
		    c.notify();
		}
	    }
	}

    }

    public void doRemoveComponent(UniqueID id) {
	WorkbenchComponent wbc = compRegistry.getWorkbenchComponent(id); 
	if (wbc != null) {
	    if (wbc instanceof WorkbenchParticipant)
	        ParticipantTool.removeParticipant(wb, 
		    (WorkbenchParticipant)wbc);
	    else {
	        compRegistry.removeWorkbenchComponent(wbc);
	        wbc.remove();
	    }
	}
    }

    public void doSetBounds(UniqueID id, Rectangle b) {
	WorkbenchComponent wbc = compRegistry.getWorkbenchComponent(id);
	if (wbc == null) 
	    System.err.println("Warning: No component with id " + id);
	else {
	    synchronized(currentBoundsLock) {
	        this.currentBounds = b;
	        wbc.getView().setBounds(b);
	        wb.componentBoundsChanged(wbc);
	        wb.repaint();
	        this.currentBounds = null;
	    }
	}
    }

    public void doMakeLink(String linkClassName, UniqueID srcID, 
	UniqueID destID, UniqueID linkID, Serializable state) {

        Class linkClass = null;
        // This is a problem. We try Class.forName first to see if the
        // system classloader can find this class. Otherwise our only
        // other option is to try to Jar loader. There are probably
        // cases where neither of these options will work. Hmmmmmm....
        try {
            linkClass = Class.forName(linkClassName);
        }
        catch(Exception e) {
            try {
                linkClass = 
		    sieve.workbench.jar.SimpleClassLoader.ourLoader.loadClass(
                    linkClassName);
            }
            catch(Exception ex) {
                ex.printStackTrace();
            }
        }

	WorkbenchComponent src = compRegistry.getWorkbenchComponent(srcID);
	WorkbenchComponent dest = compRegistry.getWorkbenchComponent(destID);

	Link link = compRegistry.getLink(linkID);

	if (src == null)
	    System.err.println("Warning: Cannot find link source: " + srcID);
	else if (dest == null)
	    System.err.println("Warning: Cannot find link dest: " + destID);
	else if (link != null)
	    System.err.println("Warning: Link already exists: " + linkID);
	else {
	    // This may be a problem because it currently will NOT
	    // cause the Workbench to notify WorkbenchListeners that
	    // a link has been added...
	    synchronized(currentLinkLock) {
		try {
		    link = (Link) linkClass.newInstance();
		    link.init(wb, src, dest, state);
	            link.setUniqueID(linkID);
	            compRegistry.registerLink(link);
	            wb.add(link.getView());
	            link.getView().setBounds(0, 0, wb.getSize().width, 
		        wb.getSize().height);
	            wb.linkAdded(link);
		}
		catch(Exception e) {
		    e.printStackTrace();
		}
	        this.currentLink = link;
	    }
	}
    }

    public void doRemoveLink(UniqueID linkID) {
	Link l = compRegistry.getLink(linkID);
	if (l != null) {
	    synchronized(currentLinkLock) {
	        this.currentLink = l;
	        compRegistry.removeLink(l);
	        l.remove(); // This should call WB.linkRemoved(...), right?
	        wb.repaint();
	        this.currentLink = null;
	    }
	}
    }

    public void doEditComponent(UniqueID objID, EditEvent e) {
	WorkbenchComponent wbc = compRegistry.getWorkbenchComponent(objID);
	if (wbc == null)
	    System.err.println("Warning: Could not find object " + objID);
	else {
	    Object obj = wbc.getModel();
	    synchronized(currentEditLock) {
		this.currentEdit = e;
		e.applyChange(obj);
		wb.componentEdited(wbc, e);
		this.currentEdit = null;
	    }
	}
    }

    /**
     * Returns the current link being created or removed by the controller.
     * The Broadcaster should call this to keep from re-broadcasting changes.
     */
    public Link getCurrentLink() {
	return currentLink;
    }
    /**
     * Returns the current property change event being applied.
     * The Broadcaster should call this to keep from re-broadcasting changes.
     */
    public EditEvent getCurrentEditEvent() {
	return currentEdit;
    }
    /**
     * Returns the bounds of the component currently being moved.
     * The Broadcaster should call this to keep from re-broadcasting changes.
     */
    public Rectangle getCurrentBounds() {
	return currentBounds;
    }

    private class Lock extends Object {}

}


