package sieve.workbench.collab;

import sieve.workbench.*;
import sieve.workbench.event.*;
import sieve.workbench.collab.msg.*;
import java.util.*;
import java.io.*;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.io.Serializable;

/**
 * MinimalReplayRecorder is a ChannelRecorder that provides support for 
 * persistance and late joiners. This recorder:
 * <OL>
 * <LI> Maintains the relative order of all recorded messages
 * <LI> Records only the most recent bounds message for each component
 * <LI> Records only the most recent property change message for each 
 *      changed property in each component component
 * <LI> Removes all messages which reference a deleted component
 * <LI> Removes link creation messages on link removal
 * <LI> Removes WorkbenchParticipants (ie, telepointers) upon
 *      deserialization
 * <LI> Removes WorkbenchParticipants who have been idle for long periods of
 *      time. (This needs further exploration.)
 * </OL>
 *
 * This should probably still track to see when there are no participants
 * left so that it can be saved to disk and garbage collected by the Server.
 *
 * @see ChannelRecorder
 */
public class MinimalReplayRecorder implements ChannelRecorder, Serializable,
    Runnable {
    /**
     * Our name (eg, the name of the channel we are recording on)
     */
    private String name = null;

    /**
     * Table mapping UniqueID's to records describing the interesting
     * messages stored for each component.
     */
    private Hashtable componentRecords = null;

    /**
     * Table mapping UniqueID's to records describing the currently active
     * links.
     */
    private Hashtable linkRecords = null;

    /**
     * Ordered vector of all relevant messages.
     */
    private Vector messages = null;

    /**
     * Thread for auditing the componentRecords table and removing any
     * old telepointers (WorkbenchParticipants)
     */
    private transient Thread auditThread = null;

    public MinimalReplayRecorder(String name) {
	messages = new Vector();
	componentRecords = new Hashtable();
	linkRecords = new Hashtable();
	this.name = name;
	startAuditThread();
    }

    private void startAuditThread() {
	auditThread = new Thread(this);
	auditThread.setDaemon(true);
	auditThread.start();
    }

    public String getName() {
	return name;
    }

    public int getMessageCount() {
	return messages.size();
    }

    public synchronized Message[] getMessages() {
	Message[] msgs = new Message[messages.size()];
	messages.copyInto(msgs);
	return msgs;
    }

    public synchronized void doAddComponent(String className, 
	MouseEvent[] evts, UniqueID id) {
	Message m = new AddComponentMessage(id, className, evts);
	componentRecords.put(id, new ComponentRecord(m, className));
	messages.addElement(m);
    }

    public synchronized void doRemoveComponent(UniqueID id) {
	ComponentRecord cr = (ComponentRecord) componentRecords.get(id);
	if (cr != null) {
	    cr.remove(messages);
	    componentRecords.remove(id);
	}
    }

    public synchronized void doSetBounds(UniqueID id, Rectangle b) {
	ComponentRecord cr = (ComponentRecord) componentRecords.get(id);
	if (cr != null) {
	    Message m = new BoundsMessage(id, b);
	    Message old = cr.resetBoundsMessage(m);
	    if (old != null)
	        messages.removeElement(old);
	    messages.addElement(m);
	}
    }
 
    public synchronized void doMakeLink(String linkClassName, UniqueID srcID, 
	UniqueID destID, UniqueID linkID, Serializable state) {
	ComponentRecord scr = (ComponentRecord) componentRecords.get(srcID);
	ComponentRecord dcr = (ComponentRecord) componentRecords.get(destID);
	if ((scr != null) && (dcr != null)) {
	    Message m = new MakeLinkMessage(linkClassName, srcID, destID, 
	    	linkID, state);
	    LinkRecord lr = new LinkRecord(scr, dcr, m, linkID);
	    scr.addLinkRecord(lr);
	    dcr.addLinkRecord(lr);
	    messages.addElement(m);
	}
    }
 
    public synchronized void doRemoveLink(UniqueID linkID) {
	Message m = new RemoveLinkMessage(linkID);
	messages.addElement(m);
    }
 
    public synchronized void doEditComponent(UniqueID objID, EditEvent e) {
	ComponentRecord cr = (ComponentRecord)componentRecords.get(objID);
	if (cr != null) {
		Message m = new EditMessage(objID, e);
		cr.registerEdit(m, e, messages);
		messages.addElement(m);
	}
    }

    private synchronized void audit() {
// System.out.println("In audit");
	for (Enumeration e = componentRecords.keys(); e.hasMoreElements();){
	    Object key = e.nextElement();
	    ComponentRecord carcass = 
		(ComponentRecord) componentRecords.get(key);
	    if (carcass.isDead()) {
// System.out.println("In audit: Removing participant carcass");
		carcass.remove(messages);
		componentRecords.remove(key);
	    }
	}
    }

    private void readObject(ObjectInputStream in) throws 
	    IOException, ClassNotFoundException {
        try {
   	     in.defaultReadObject();
        }
        catch(NotActiveException e) {
	    e.printStackTrace();
	    return;
	}
	audit();
	startAuditThread();
    }

    public synchronized void run() {
	try {
	    while(true) {
		wait(5 * 60 * 1000);
		audit();
	    }
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }

    /**
     * There's one ComponentRecord for each component in the system. It
     * tracks the current set of "interesting" messages for that component
     * so that they can be removed if needed.
     */
    private class ComponentRecord implements Serializable {
        private Vector edits = null;
	private Vector linkRecords = null;
	private Message boundsMessage = null; 
        private Message createMessage = null;
	private String componentClass = null;
	private transient boolean dead = false;
	private transient boolean isKillable = false;
	private transient long lastModification = 0;

	public ComponentRecord(Message createMessage, String componentClass) {
	    linkRecords = new Vector();
	    edits = new Vector();
	    this.createMessage = createMessage;
	    this.componentClass = componentClass;
	    isKillable =
		componentClass.equals(WorkbenchParticipant.class.getName());
	    if (isKillable)
	        lastModification = System.currentTimeMillis();
	}

	public void remove(Vector messages) {
	    // Remove the message that created us
	    if (createMessage != null)
	        messages.removeElement(createMessage);
	    createMessage = null;

	    // Remove our bounds message
	    if (boundsMessage != null)
	        messages.removeElement(boundsMessage);
	    boundsMessage = null;

	    // Remove all of our edit messages
	    for (Enumeration e = edits.elements(); e.hasMoreElements();)
		messages.removeElement((Message)e.nextElement());
	    edits.removeAllElements();
	    edits = null;

	    // Remove all links to or from us
	    for (Enumeration e = linkRecords.elements(); e.hasMoreElements();)
		((LinkRecord)e.nextElement()).remove(messages);
	    linkRecords.removeAllElements();
	    linkRecords = null;
	}

	public boolean isDead() {
	    return dead || (isKillable && 
	      (lastModification - System.currentTimeMillis()) > 10 * 60 * 1000);
	}

	public Message resetBoundsMessage(Message m) {
	    Message tmp = boundsMessage;
	    boundsMessage = m;
	    if (isKillable)
	        lastModification = System.currentTimeMillis();
	    return tmp;
	}

	/**
	 * Checks this edit against all previous edits for this component.
	 * Any previous edits which are obsoleted by this message are
	 * deleted.
         */
	public void registerEdit(Message m, EditEvent ee, Vector msgs) {
	    Vector editsClone = (Vector) edits.clone();
	    for (Enumeration e = editsClone.elements(); e.hasMoreElements();) {
		EditMessage old = (EditMessage) e.nextElement();
		if (old.getEditEvent().isObsoletedBy(ee)) {
		    edits.removeElement(old);
		    msgs.removeElement(old);
		}
	    }
	    edits.addElement(m);

	    if (isKillable)
	        lastModification = System.currentTimeMillis();
	}

	public void removeLinkRecord(LinkRecord lr) {
	    linkRecords.removeElement(lr);
	}

	public void addLinkRecord(LinkRecord lr) {
	    linkRecords.addElement(lr);
	}

	private void readObject(ObjectInputStream in) throws 
	    IOException, ClassNotFoundException {
	    try {
	        in.defaultReadObject();
	    }
	    catch(NotActiveException e) {
		e.printStackTrace();
		return;
	    }
	    // If we are a WorkbenchParticipant (i.e., telepointer) and
	    // are being read off of the disk, then we are obviously no
	    // longer alive.
	    dead = 
		componentClass.equals(WorkbenchParticipant.class.getName());
	}

    }

    /**
     * A LinkRecord is a record of a link message.
     */
    private class LinkRecord implements Serializable {
	private ComponentRecord srcRec = null;
	private ComponentRecord destRec = null;
	private Message m = null;
	private UniqueID id = null;

	public LinkRecord(ComponentRecord srcRec, ComponentRecord destRec,
	    Message m, UniqueID id) {
	    this.srcRec = srcRec;
	    this.destRec = destRec;
	    this.m = m;
	    this.id = id;
	    linkRecords.put(id, this);
	}

	public void remove(Vector messages) {
	    messages.removeElement(m);
	    srcRec.removeLinkRecord(this);
	    destRec.removeLinkRecord(this);
	    linkRecords.remove(id);
	}
    }

}



