package sieve.workbench.collab;

import java.util.Hashtable;
import java.io.Serializable;
import java.awt.*;
import java.awt.event.*;
import sieve.workbench.*;
import sieve.workbench.event.*;

/**
 * A workbench controller for replaying messages from a file. This ReplayController
 * was written to be used by the SingleUserSupport class which allows the sieve
 * workspace user to save and restore sessions locally. Messages in the serialized 
 * file have UniqueIDs associated with them. In order to load multiple copies of
 * a saved collection of components, new UniqueIDs are assigned to be used by the
 * workbench. However, within the loading file, there will be multiple messages
 * referencing a single component. The ReplayController uses a hash table to map
 * the UniqueIDs in the serialized file to the newly created IDs. CAUTION: the 
 * UniqueID constructor uses the Object.getClass().getName() in the identifying
 * string. Here we are passing in for every type of component a String object whose
 * value is the class name. Thus the class name portion of the UniqueID will
 * always be "String" and an better arrangement will be necessary for any functionality
 * which depends on the identified component's actual class name.
 * 
 * @version	%I%, %G%
 * @author	Lynn W. Jones
 *
 */

public class ReplayController implements WorkbenchController {
	private Hashtable mapper = new Hashtable();
	private Controller wbController = null;

	public ReplayController(Controller wbc) {
		wbController = wbc;
	}

	//I am assuming here that we only need to generate unique ID's for objects
	//in adding them and making links. Other operations are meaningless unless
	//the objects already exist, so their ID's will be retrieved from the hashtable

    public void doAddComponent(String className, MouseEvent[] evts, UniqueID id) {
		UniqueID newID = new UniqueID((Object)className);	//get a new id based on the class name
		mapper.put(id, newID);	//put the new ID value at the hash location of the old one
		wbController.doAddComponent(className, evts, newID);
	}

	public void doRemoveComponent(UniqueID id){
		wbController.doRemoveComponent((UniqueID)mapper.get(id));	
	}

    public void doSetBounds(UniqueID id, Rectangle b){
		wbController.doSetBounds((UniqueID)mapper.get(id), b);	
	}

    public void doMakeLink(String linkClassName, UniqueID srcID, UniqueID destID, UniqueID linkID, Serializable state) {
		UniqueID newLinkID = new UniqueID((Object)linkClassName);	//get a new id based on the link class name
		mapper.put(linkID, newLinkID);	//put the new ID value at the hash location of the old one
		UniqueID source = (UniqueID)mapper.get(srcID);	//the source and destination must already be in the table
		UniqueID dest = (UniqueID)mapper.get(destID);
		wbController.doMakeLink(linkClassName, source, dest, newLinkID, state);
	}

    public void doRemoveLink(UniqueID linkID) {
		wbController.doRemoveLink((UniqueID)mapper.get(linkID));	
	}

    public void doEditComponent(UniqueID objID, EditEvent e) {
		wbController.doEditComponent((UniqueID)mapper.get(objID), e);	
	}

}
